#!/usr/bin/env python3
"""OCS agent for thermometer board in the stimulator box
"""
import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.stimulator_thermometer.drivers import (Max31856,
                                                        from_spi_node_path)

LOCK_RELEASE_SEC = 1.
LOCK_RELEASE_TIMEOUT = 10


class StimThermometerAgent:
    """OCS agent class for stimulator thermometer.

    Parameters
    ----------
    paths_spinode : list of str or pathlib.Path
        List of path to SPI nodes of temperature devices.
    """

    def __init__(self, agent, paths_spinode):
        self.active = True
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.take_data = False

        self.devices = [from_spi_node_path(path) for path in paths_spinode]

        self.initialized = False

        agg_params = {'frame_length': 60}
        self.agent.register_feed('temperatures',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('sampling_frequency', type=float, default=1)
    def acq(self, session, params):
        """acq()

        **Process** - Fetch temperatures from the thermometer readers.

        Parameters
        ----------
        sampling_frequency : float, optional
            Sampling frequency in Hz, default to 1 Hz.

        Notes
        -----
        An example of the session data::

            >>> response.session['data']

            {"Channel_0": {"T":15.092116135817285},
             "Channel_1": {"T":15.092116135817285},
             "Channel_2": {"T":15.40625,"T_cj":16.796875},
             "Channel_3": {"T":15.4921875,"T_cj":16.453125}},
             ...
             "timestamp": 1738827824.2523127
            }
        """
        f_sample = params['sampling_frequency']
        pace_maker = Pacemaker(f_sample)

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn(
                    f'Could not start acq because {self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            self.take_data = True
            session.data = {}
            last_release = time.time()

            while self.take_data:
                if time.time() - last_release > LOCK_RELEASE_SEC:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=LOCK_RELEASE_TIMEOUT):
                        print(f'Re-acquire failed: {self.lock.job}')
                        return False, 'Could not re-acquire lock.'

                current_time = time.time()

                for ch_num, dev in enumerate(self.devices):
                    data = {'timestamp': current_time, 'block_name': f'temp_{ch_num}', 'data': {}}
                    chan_string = f'Channel_{ch_num}'

                    temp = dev.get_temp()
                    data['data'][chan_string + '_T'] = temp

                    if isinstance(dev, Max31856):
                        cjtemp = dev.get_temp_ambient()
                        data['data'][chan_string + '_T_cj'] = cjtemp
                        field_dict = {chan_string: {'T': temp, 'T_cj': cjtemp}}
                    else:
                        field_dict = {chan_string: {'T': temp}}

                    self.agent.publish_to_feed('temperatures', data)
                    session.data.update(field_dict)

                session.data.update({'timestamp': current_time})
                pace_maker.sleep()

            self.agent.feeds['temperatures'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'

        return False, 'acq is not currently running.'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--paths-spinode', nargs='*', type=str, required=True,
                        help="Path to the spi nodes.")

    return parser


def main(args=None):
    """Boot OCS agent"""
    txaio.start_logging(level=os.environ.get('LOGLEVEL', 'info'))

    parser = make_parser()
    args = site_config.parse_args('StimThermometerAgent',
                                  parser=parser,
                                  args=args)

    agent_inst, runner = ocs_agent.init_site_agent(args)
    stim_max_agent = StimThermometerAgent(agent_inst, args.paths_spinode)

    agent_inst.register_process(
        'acq',
        stim_max_agent.acq,
        stim_max_agent._stop_acq,
        startup=True
    )

    runner.run(agent_inst, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
from os import environ

import requests
import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock


class UCSCRadiometerAgent:
    """Monitor the PWV Flask Server.

    Parameters
    ----------
    agent : OCS Agent
        OCSAgent object which forms this Agent
    url : str
        url of the radiometer web server on the internet

    """

    def __init__(self, agent, url):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.url = url

        self.take_data = False

        agg_params = {'frame_length': 60,
                      'exclude_influx': False}

        # register the feed
        self.agent.register_feed('pwvs',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1
                                 )

        self.last_published_reading = None

    @ocs_agent.param('test_mode', default=False, type=bool)
    def acq(self, session, params=None):
        """acq()

        **Process** - Fetch values from PWV Flask Server

        Parameters
        ----------
        test_mode : bool, option
            Run the Process loop only once. Meant only for testing.
            Default is False.

        Notes
        -----
        The most recent data collected is stored in session data in the
        following structure::

            >>> response.session['data']
            {'timestamp': 1678820419.0,
             'pwv': 0.49253026985972237}

        """
        pm = Pacemaker(1 / 60, quantize=False)

        self.take_data = True

        while self.take_data:
            try:
                r = requests.get(self.url, timeout=60)
            except ValueError:
                pm.sleep()
                continue
            except requests.exceptions.ConnectionError as e:
                self.log.warn(f"Connection error occured: {e}")
                pm.sleep()
                continue
            except requests.exceptions.Timeout as e:
                self.log.warn(f"Timeout exception occurred: {e}")
                pm.sleep()
                continue
            data = r.json()
            last_pwv = data['pwv']
            last_timestamp = data['timestamp']

            pwvs = {'block_name': 'pwvs',
                    'timestamp': last_timestamp,
                    'data': {'pwv': last_pwv}
                    }

            if self.last_published_reading is not None:
                if last_timestamp > self.last_published_reading[1]:
                    self.agent.publish_to_feed('pwvs', pwvs)
                    self.last_published_reading = (last_pwv, last_timestamp)
            else:
                self.agent.publish_to_feed('pwvs', pwvs)
                self.last_published_reading = (last_pwv, last_timestamp)

            session.data = {"timestamp": last_timestamp,
                            "pwv": last_pwv}

            if params['test_mode']:
                break
            else:
                pm.sleep()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """
        self.take_data = False
        return True, 'Stopping acq process'


def add_agent_args(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument("--url", type=str, help="url for radiometer web server")
    pgroup.add_argument("--test-mode", action='store_true',
                        help="Determines whether agent runs in test mode."
                        "Default is False.")
    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = add_agent_args()
    args = site_config.parse_args(agent_class='UCSCRadiometerAgent', parser=parser, args=args)

    # test params
    test_params = {'test_mode': False}
    if args.test_mode:
        test_params = {'test_mode': True}

    agent, runner = ocs_agent.init_site_agent(args)
    pwv_agent = UCSCRadiometerAgent(agent, args.url)

    agent.register_process('acq', pwv_agent.acq, pwv_agent._stop_acq, startup=test_params)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



# DWL drivers
from socs.agents.wiregrid_tiltsensor.drivers.dwl import DWL
# sherborne drivers
from socs.agents.wiregrid_tiltsensor.drivers.sherborne import Sherborne


def connect(ip, port, sensor_type):
    if sensor_type == 'DWL':
        tiltsensor = DWL(tcp_ip=ip, tcp_port=port, timeout=0.5, isSingle=False, verbose=0)
    elif sensor_type == 'sherborne':
        tiltsensor = Sherborne(tcp_ip=ip, tcp_port=port, reset_boot=False, timeout=0.5, verbose=0)
    else:
        raise ('Invalid tiltsensor type')
    return tiltsensor



# This device uses the Prologix GPIB interface
from socs.common.prologix_interface import PrologixInterface


class SRSCG635Interface(PrologixInterface):
    """
    This device driver is written for the SRS CG635 clock used for the timing system.
    """

    def __init__(self, ip_address, gpibAddr, verbose=False, **kwargs):
        self.verbose = verbose
        super().__init__(ip_address, gpibAddr, **kwargs)

    def get_freq(self):
        """
        Queries the clock for its current output frequency in Hz.

        Returns the frequency as a float.
        """

        self.write("FREQ?")
        freq = self.read()

        return float(freq)

    def get_stdc(self):
        """
        Queries the clock for the current Standard CMOS (STDC) output setting.

        The query returns an int with the int representing the CMOS output setting.
        The outputs are represented in volts between the CMOS low and CMOS high with CMOS low = 0V.

        The standard CMOS output settings this query can return are are::

            -1 = Not a standard CMOS Output
             0 = 1.2V
             1 = 1.8V
             2 = 2.5V
             3 = 3.3V (The default for our current setup)
             4 = 5.0V

        """

        self.write("STDC?")
        stdc = self.read()

        return int(stdc)

    def get_runs(self):
        """
        Queries the clock for the current Running State (RUNS).

        Returns an int which represents the following running states::

            0 = Not Running (Output is off)
            1 = Running (Output is on)

        """

        self.write("RUNS?")
        runs = self.read()

        return int(runs)

    def get_timebase(self):
        """
        Queries the clock for the current timebase (TIMB).

        Returns an int which represents the following states::

            0 = Internal timebase
            1 = OCXO timebase
            2 = Rubidium timebase
            3 = External timebase

        """

        self.write("TIMB?")
        timb = self.read()

        return int(timb)

    def get_all_status(self):
        self.write("FREQ?;STDC?;RUNS?;TIMB?")
        freq, stdc, runs, timb = self.read().split(';')

        return float(freq), int(stdc), int(runs), int(timb)

    def clear(self):
        """Clear all the event registers and error queue."""
        self.write("*CLS")
        return True



import argparse
import queue
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from twisted.internet import defer, reactor, threads

txaio.use_twisted()


from dataclasses import dataclass

import socs.agents.hwp_pid.drivers.pid_controller as pd


def parse_action_result(res):
    """
    Parses the result of an action to ensure it is a dictionary so it can be
    stored in session.data
    """
    if res is None:
        return {}
    elif isinstance(res, dict):
        return res
    else:
        return {'result': res}


def get_pid_state(pid: pd.PID):
    state_func = {'current_freq': pid.get_freq,
                  'target_freq': pid.get_target,
                  'direction': pid.get_direction}

    return_dict = {'healthy': True}
    for name, func in state_func.items():
        resp = func()
        if resp.msg_type == 'error':
            return_dict['healthy'] = False
        else:
            return_dict[name] = resp.measure
    return return_dict


class Actions:
    @dataclass
    class BaseAction:
        def __post_init__(self):
            self.deferred = defer.Deferred()
            self.log = txaio.make_logger()

    @dataclass
    class TuneStop(BaseAction):
        def process(self, pid: pd.PID):
            pid.tune_stop()

    @dataclass
    class TuneFreq(BaseAction):
        def process(self, pid: pd.PID):
            pid.tune_freq()

    @dataclass
    class DeclareFreq(BaseAction):
        freq: float

        def process(self, pid: pd.PID):
            pid.declare_freq(self.freq)
            return {"declared_freq": self.freq}

    @dataclass
    class SetPID(BaseAction):
        p: float
        i: int
        d: float

        def process(self, pid: pd.PID):
            pid.set_pid([self.p, self.i, self.d])

    @dataclass
    class SetDirection(BaseAction):
        direction: str

        def process(self, pid: pd.PID):
            pid.set_direction(self.direction)

    @dataclass
    class SetScale(BaseAction):
        slope: float
        offset: float

        def process(self, pid: pd.PID):
            pid.set_scale(self.slope, self.offset)

    @dataclass
    class GetState(BaseAction):
        def process(self, pid: pd.PID):
            pid_state = get_pid_state(pid)
            if pid_state['healthy']:
                return pid_state
            else:
                print('Error getting state')
                raise ValueError


class HWPPIDAgent:
    """Agent to PID control the rotation speed of the CHWP

    Args:
        ip (str): IP address for the PID controller
        port (str): Port for the PID controller
        verbosity (str): Verbosity of PID controller output

    """

    def __init__(self, agent, ip, port, verbosity):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self._initialized = False
        self.take_data = False
        self.ip = ip
        self.port = port
        self._verbosity = verbosity > 0
        self.action_queue = queue.Queue()

        agg_params = {"frame_length": 60}
        self.agent.register_feed("hwppid", record=True, agg_params=agg_params)

    def _get_data_and_publish(self, pid: pd.PID, session: ocs_agent.OpSession):
        data = {"timestamp": time.time(), "block_name": "HWPPID", "data": {}}

        pid_state = get_pid_state(pid)
        if pid_state['healthy']:
            session.degraded = False
        else:
            print('Warning: state monitor degraded')
            session.degraded = True

        data['data'].update(pid_state)
        session.data.update(pid_state)
        session.data['last_updated'] = time.time()
        self.agent.publish_to_feed("hwppid", data)

    def _process_actions(self, pid):
        while not self.action_queue.empty():
            action = self.action_queue.get()
            try:
                self.log.info(f"Running action {action}")
                res = action.process(pid)
                threads.blockingCallFromThread(
                    reactor, action.deferred.callback, res
                )
            except Exception as e:
                self.log.error(f"Error processing action: {action}")
                threads.blockingCallFromThread(
                    reactor, action.deferred.errback, e
                )

    def _clear_queue(self):
        while not self.action_queue.empty():
            action = self.action_queue.get()
            action.deferred.errback(Exception("Action cancelled"))

    def main(self, session, params):
        """main()

        **Process** - Main Process for PID agent. Periodically queries PID
            controller for data, and executes requested actions.

        Notes:
            The most recent data collected is stored in the session data in the
            structure::

                >>> response.session['data']
                {'current_freq': 0,
                 'target_freq': 0,
                 'direction': 1,
                 'last_updated': 1649085992.719602}
        """
        pid = pd.PID(ip=self.ip, port=self.port, verb=self._verbosity)
        self.log.info("Connected to PID controller")

        self._clear_queue()

        sample_period = 5.0
        last_sample = 0.0
        session.set_status("running")
        while session.status in ["starting", "running"]:
            now = time.time()
            if now - last_sample > sample_period:
                self._get_data_and_publish(pid, session)
                last_sample = now

            self._process_actions(pid)
            time.sleep(0.2)

        return True, "Exited main process"

    def _main_stop(self, session, params):
        """Stop main process"""
        session.set_status("stopping")
        return True, "Set main status to stopping"

    @defer.inlineCallbacks
    def tune_stop(self, session, params):
        """tune_stop()

        **Task** - Reverse the drive direction of the PID controller and
        optimize the PID parameters for deceleration.

        """
        action = Actions.TuneStop(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"

    @defer.inlineCallbacks
    def tune_freq(self, session, params):
        """tune_freq()

        **Task** - Tune the PID controller setpoint to the rotation frequency
        and optimize the PID parameters for rotation.

        """
        action = Actions.TuneFreq(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"

    @defer.inlineCallbacks
    @ocs_agent.param("freq", default=0.0, check=lambda x: 0.0 <= x <= 3.0)
    def declare_freq(self, session, params):
        """declare_freq(freq=0)

        **Task** - Store the entered frequency as the PID setpoint when
        ``tune_freq()`` is next called.

        Parameters:
            freq (float): Desired HWP rotation frequency

        Notes:
            Session data is structured as follows::

                >>> response.session['data']
                {'declared_freq': 2.0}
        """
        action = Actions.DeclareFreq(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"

    @defer.inlineCallbacks
    @ocs_agent.param("p", default=0.2, type=float, check=lambda x: 0.0 < x <= 8.0)
    @ocs_agent.param("i", default=63, type=int, check=lambda x: 0 <= x <= 200)
    @ocs_agent.param("d", default=0.0, type=float, check=lambda x: 0.0 <= x < 10.0)
    def set_pid(self, session, params):
        """set_pid(p=0.2, i=63, d=0.)

        **Task** - Set the PID parameters. Note these changes are for the
        current session only and will change whenever the agent container is
        reloaded.

        Parameters:
            p (float): Proportional PID value
            i (int): Integral PID value
            d (float): Derivative PID value
        """
        action = Actions.SetPID(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"

    @defer.inlineCallbacks
    @ocs_agent.param("direction", type=str, default="0", choices=["0", "1"])
    def set_direction(self, session, params):
        """set_direction(direction='0')

        **Task** - Set the HWP rotation direction.

        Parameters:
            direction (str): '0' for forward and '1' for reverse.

        """
        action = Actions.SetDirection(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"

    @defer.inlineCallbacks
    @ocs_agent.param("slope", default=1.0, type=float, check=lambda x: -10.0 < x < 10.0)
    @ocs_agent.param(
        "offset", default=0.1, type=float, check=lambda x: -10.0 < x < 10.0
    )
    def set_scale(self, session, params):
        """set_scale(slope=1, offset=0.1)

        **Task** - Set the PID's internal conversion from input voltage to
        rotation frequency.

        Parameters:
            slope (float): Slope of the "rotation frequency vs input voltage"
                relationship
            offset (float): y-intercept of the "rotation frequency vs input
                voltage" relationship

        """
        action = Actions.SetScale(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"

    @defer.inlineCallbacks
    def get_state(self, session, params):
        """get_state()

        **Task** - Polls hardware for the current the PID state.

        Notes:
            Session data for this operation is as follows::

                >>> response.session['data']
                {'current_freq': 0,
                 'target_freq': 0,
                 'direction': 1}
        """
        action = Actions.GetState(**params)
        self.action_queue.put(action)
        res = yield action.deferred
        session.data = parse_action_result(res)
        return True, f"Completed: {str(action)}"


def make_parser(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically build documentation
    baised on this function
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent
    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--ip")
    pgroup.add_argument("--port")
    pgroup.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="PID Controller verbosity level.",
    )
    pgroup.add_argument("--mode")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class="HWPPIDAgent", parser=parser, args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    if args.mode is not None:
        agent.log.warn("--mode agrument is deprecated.")

    hwppid_agent = HWPPIDAgent(
        agent, ip=args.ip, port=args.port, verbosity=args.verbose
    )
    agent.register_process(
        "main", hwppid_agent.main, hwppid_agent._main_stop, startup=True
    )
    agent.register_task("tune_stop", hwppid_agent.tune_stop, blocking=False)
    agent.register_task("tune_freq", hwppid_agent.tune_freq, blocking=False)
    agent.register_task("declare_freq", hwppid_agent.declare_freq, blocking=False)
    agent.register_task("set_pid", hwppid_agent.set_pid, blocking=False)
    agent.register_task("set_direction", hwppid_agent.set_direction, blocking=False)
    agent.register_task("set_scale", hwppid_agent.set_scale, blocking=False)
    agent.register_task("get_state", hwppid_agent.get_state, blocking=False)
    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



# Script to log and readout PTC data through ethernet connection.
# Tamar Ervin and Jake Spisak, February 2019
# Sanah Bhimani, May 2022

import argparse
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.agents.cryomech_cpa.drivers import PTC


class PTCAgent:
    """Agent to connect to a single cryomech compressor.

    Parameters:
        port (int): TCP port to connect to.
        ip_address (str): IP Address for the compressor.
        f_sample (float, optional): Data acquisiton rate, defaults to 2.5 Hz.
        fake_errors (bool, optional): Generates fake errors in the string
            output 50% of the time.

    """

    def __init__(self, agent, port, ip_address, f_sample=2.5,
                 fake_errors=False):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.ip_address = ip_address
        self.fake_errors = fake_errors

        self.port = port
        self.module = None
        self.f_sample = f_sample

        self.initialized = False
        self.take_data = False

        # Registers data feeds
        agg_params = {
            'frame_length': 60,
        }
        self.agent.register_feed('ptc_status',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init(self, session, params=None):
        """init(auto_acquire=False)

        **Task** - Initializes the connection to the PTC.

        Parameters:
            auto_acquire (bool): Automatically start acq process after
                initialization if True. Defaults to False.

        """
        if self.initialized:
            return True, "Already Initialized"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            # Establish connection to ptc
            self.ptc = PTC(self.ip_address, port=self.port,
                           fake_errors=self.fake_errors)

            # Test connection and display identifying info
            try:
                self.ptc.get_data()
            except ConnectionError:
                self.log.error("Could not establish connection to compressor.")
                return False, "PTC agent initialization failed"
            print("PTC Model:", self.ptc.model)
            print("PTC Serial Number:", self.ptc.serial)
            print("Software Revision is:", self.ptc.software_revision)

        self.initialized = True

        # Start data acquisition if requested
        if params['auto_acquire']:
            resp = self.agent.start('acq', params={})
            self.log.info(f'Response from acq.start(): {resp[1]}')

        return True, "PTC agent initialized"

    @ocs_agent.param('state', type=str, choices=['off', 'on'])
    def power_ptc(self, session, params=None):
        """power_ptc(state=None)

        **Task** - Remotely turn the PTC on or off.

        Parameters
        ----------
        state : str
            Desired power state of the PTC, either 'on', or 'off'.

        """
        with self.lock.acquire_timeout(3, job='power_ptc') as acquired:
            if not acquired:
                self.log.warn("Could not start task because {} is already "
                              "running".format(self.lock.job))
                return False, "Could not acquire lock."

            self.ptc.power(params['state'])

        return True, "PTC powered {}".format(params['state'])

    @ocs_agent.param('test_mode', default=False, type=bool)
    def acq(self, session, params):
        """acq()

        **Process** - Starts acqusition of data from the PTC.

        Parameters:
            test_mode (bool, optional): Run the Process loop only once.
                This is meant only for testing. Default is False.

        """
        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already"
                              "running".format(self.lock.job))
                return False, "Could not acquire lock."

            last_release = time.time()

            self.take_data = True

            while self.take_data:
                # Relinquish sampling lock occasionally
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire sampling lock, "
                                      f"currently held by {self.lock.job}.")
                        continue

                # Publish data, waiting 1/f_sample seconds in between calls.
                pub_data = {'timestamp': time.time(),
                            'block_name': 'ptc_status'}
                try:
                    data_flag, data = self.ptc.get_data()
                    if session.degraded:
                        self.log.info("Connection re-established.")
                        session.degraded = False
                except ConnectionError:
                    self.log.error("Failed to get data from compressor. Check network connection.")
                    session.degraded = True
                    time.sleep(1)
                    continue
                pub_data['data'] = data
                # If there is an error in compressor output (data_flag = True),
                # do not publish
                if not data_flag:
                    self.agent.publish_to_feed('ptc_status', pub_data)
                time.sleep(1. / self.f_sample)

                if params['test_mode']:
                    break

            self.agent.feeds["ptc_status"].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params):
        """Stops acqusition of data from the PTC."""
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--port', default=502)
    pgroup.add_argument('--serial-number')
    pgroup.add_argument('--mode', choices=['init', 'acq'])
    pgroup.add_argument('--fake-errors', default=False,
                        help="If True, randomly output 'FAKE ERROR' instead of "
                             "data half of the time.")

    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='CryomechCPAAgent',
                                  parser=parser,
                                  args=args)
    print('I am in charge of device with serial number: %s' % args.serial_number)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    # Call launcher function (initiates connection to appropriate
    # WAMP hub and realm).

    agent, runner = ocs_agent.init_site_agent(args)

    # create agent instance and run log creation
    ptc = PTCAgent(agent, args.port, args.ip_address,
                   fake_errors=args.fake_errors)

    agent.register_task('init', ptc.init, startup=init_params)
    agent.register_process('acq', ptc.acq, ptc._stop_acq)
    agent.register_task('power_ptc', ptc.power_ptc)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import random
import struct

from socs.tcp import TCPInterface

STX = '\x02'
ADDR = '\x10'
CMD = '\x80'
CR = '\x0D'
DATA_WRITE = '\x61'
DATA_READ = '\x63'
ESC = '\x07'
ESC_STX = '\x30'
ESC_CR = '\x31'
ESC_ESC = '\x32'


class PTC(TCPInterface):
    """Interface class for connecting to the pulse tube compressor.

    Parameters
    ----------
    ip_address : str
        IP address of the device.
    port : int
        Associated port for TCP communication. Default is 502.
    timeout : float
        Duration in seconds that operations wait before giving up. Default is
        10 seconds.
    fake_errors : bool
        Flag that generates random fake errors if True. Does not generate
        errors if False. Defaults to False.

    Attributes
    ----------
    comm : socket.socket
        Socket object that forms the connection to the compressor.

    """

    def __init__(self, ip_address, port=502, timeout=10, fake_errors=False):
        self.fake_errors = fake_errors

        self.model = None
        self.serial = None
        self.software_revision = None

        # Setup the TCP Interface
        super().__init__(ip_address, port, timeout)

    def get_data(self):
        """
        Gets the raw data from the PTC and returns it in a usable format.
        """
        self.send(self.buildRegistersQuery())
        data = self.recv(1024)
        data_flag, brd = self.breakdownReplyData(data)

        return data_flag, brd

    @staticmethod
    def buildRegistersQuery():
        query = bytes([0x09, 0x99,  # Message ID
                       0x00, 0x00,  # Unused
                       0x00, 0x06,  # Message size in bytes
                       0x01,        # Slave Address
                       0x04,        # Function Code
                       0x00, 0x01,   # The starting Register Number
                       0x00, 0x35])  # How many to read
        return query

    def power(self, state):
        """Turn the PTC on or off.

        Parameters
        ----------
        state : str
            Desired power state of the PTC, either 'on', or 'off'.

        """
        command = [0x09, 0x99,  # Message ID
                   0x00, 0x00,  # Unused
                   0x00, 0x06,  # Message size in bytes
                   0x01,        # Slave Address
                   0x06,        # Function Code
                   0x00, 0x01]   # Register Number

        if state.lower() == 'on':
            command.extend([0x00, 0x01])
        elif state.lower() == 'off':
            command.extend([0x00, 0xff])
        else:
            raise ValueError(f"Invalid state: {state}")

        self.comm.sendall(bytes(command))
        self.comm.recv(1024)  # Discard the echoed command

    def breakdownReplyData(self, rawdata):
        """Take in raw ptc data, and return a dictionary.

        The dictionary keys are the data labels, the dictionary values are the
        data in floats or ints.

        Returns
        -------
        data_flag : bool
            False if data is valid, True if output could not be interpretted.
        data : dict
            Data dictionary already formatted for passing to OCS Feed.

        """

        # Associations between keys and their location in rawData
        keyloc = {"Operating_State": [9, 10],
                  "Compressor_State": [11, 12],
                  "Warning_State": [15, 16, 13, 14],
                  "Alarm_State": [19, 20, 17, 18],
                  "Coolant_In_Temp": [23, 24, 21, 22],
                  "Coolant_Out_Temp": [27, 28, 25, 26],
                  "Oil_Temp": [31, 32, 29, 30],
                  "Helium_Temp": [35, 36, 33, 34],
                  "Low_Pressure": [39, 40, 37, 38],
                  "Low_Pressure_Average": [43, 44, 41, 42],
                  "High_Pressure": [47, 48, 45, 46],
                  "High_Pressure_Average": [51, 52, 49, 50],
                  "Delta_Pressure_Average": [55, 56, 53, 54],
                  "Motor_Current": [59, 60, 57, 58],
                  "Hours_of_Operation": [63, 64, 61, 62],
                  "Pressure_Unit": [65, 66],
                  "Temperature_Unit": [67, 68],
                  "Serial_Number": [69, 70],
                  "Model": [71, 72],
                  "Software_Revision": [73, 74]}

        # Iterate through all keys and return the data in a usable format.
        # If there is an error in the string format, print the
        # error to logs, return an empty dictionary, and flag the data as bad
        data = {}

        # If fake_errors=True, then randomly output the string 'FAKE ERROR'
        # instead of the actual data 50% of the time
        if self.fake_errors:
            if random.random() < 0.5:
                rawdata = "FAKE ERROR"

        try:
            for key in keyloc.keys():
                locs = keyloc[key]
                wkrBytes = bytes([rawdata[loc] for loc in locs])

                # four different data formats to unpack
                # Big endian unsigned integer 16 bits
                if key in [
                    "Operating_State",
                    "Compressor_State",
                    "Pressure_Unit",
                    "Temperature_Unit",
                    "Serial_Number",
                ]:
                    state = struct.unpack(">H", wkrBytes)[0]
                    # Serial number is an attribute, not publishable data
                    if key == "Serial_Number":
                        self.serial = state
                    else:
                        data[key] = state
                # 32bit signed integer which is actually stored as a
                # 32bit IEEE float (silly)
                elif key in ["Warning_State", "Alarm_State"]:
                    state = int(struct.unpack(">f", wkrBytes)[0])
                    data[key] = state
                # 2 x 8-bit lookup tables.
                elif key in ["Model"]:
                    model_major = struct.unpack(
                        ">B", bytes([rawdata[locs[0]]]))[0]
                    model_minor = struct.unpack(
                        ">B", bytes([rawdata[locs[1]]]))[0]
                    # Model is an attribute, not publishable data
                    self.model = str(model_major) + "_" + str(model_minor)
                elif key in ["Software_Revision"]:
                    version_major = struct.unpack(
                        ">B", bytes([rawdata[locs[0]]]))[0]
                    version_minor = struct.unpack(
                        ">B", bytes([rawdata[locs[1]]]))[0]
                    self.software_revision = str(version_major) + "." + str(version_minor)
                # 32 bit Big endian IEEE floating point
                else:
                    data[key] = struct.unpack(">f", wkrBytes)[0]

            data_flag = False

        except BaseException:
            data_flag = True
            print("Compressor output could not be converted to numbers."
                  f"Skipping this data block. Bad output string is {rawdata}")

        return data_flag, data



import argparse
import os
import time

import txaio
from autobahn.twisted.util import sleep as dsleep
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from twisted.internet.defer import inlineCallbacks

from socs.snmp import SNMPTwister

# For logging
txaio.use_twisted()

CONVERT_OIDS = ['mbgSyncboxN2XPtpUTCOffset',
                'mbgSyncboxN2XPtpOffsetToGrandmaster',
                'mbgSyncboxN2XPtpMeanPathDelay']


def _convert_oid_value(field_name, oid_value):
    unit_multipliers = {
        "s": 1000000,    # seconds to microseconds
        "sec": 1000000,  # seconds to microseconds
        "ms": 1000,      # milliseconds to microseconds
        "us": 1,         # microseconds
        "ns": 0.001      # nanoseconds to microseconds
    }

    value, unit = oid_value.split()
    value = float(value)

    # Normalize to microseconds
    if unit in unit_multipliers:
        oid_value = value * unit_multipliers[unit]

    # Change the field name to differentiate from raw output
    field_name = field_name + "_value"

    return field_name, oid_value


def _extract_oid_field_and_value(get_result):
    """Extract field names and OID values from SNMP GET results.

    The ObjectType objects returned from pysnmp interactions contain the
    info we want to use for field names, specifically the OID and associated
    integer for uniquely identifying duplicate OIDs, as well as the value of the
    OID, which we want to save.

    Here we use the prettyPrint() method to get the OID name, requiring
    some string manipulation. We also just grab the hidden ._value
    directly, as this seemed the cleanest way to get an actual value of a
    normal type. Without doing this we get a pysnmp defined Integer32 or
    DisplayString, which were awkward to handle, particularly the
    DisplayString.

    Parameters
    ----------
    get_result : pysnmp.smi.rfc1902.ObjectType
        Result from a pysnmp GET command.

    Returns
    -------
    field_name : str
        Field name for an OID, i.e. 'outletStatus_1'
    oid_value : int or str
        Associated value for the OID. Returns None if not an int or str
    oid_description : str
        String description of the OID value.
    """
    # OID from SNMP GET
    oid = get_result[0].prettyPrint()
    # Makes something like 'IBOOTPDU-MIB::outletStatus.1'
    # look like 'outletStatus_1'
    field_name = oid.split("::")[1].replace('.', '_')

    # Grab OID value, mostly these are integers
    oid_value = get_result[1]._value
    oid_description = get_result[1].prettyPrint()

    # Decode string values
    if isinstance(oid_value, bytes):
        oid_value = oid_value.decode("utf-8")

    # I don't expect any other types at the moment, but just in case.
    if not isinstance(oid_value, (int, bytes, str)):
        oid_value = None

    return field_name, oid_value, oid_description


def _build_message(get_result, time):
    """Build the message for publication on an OCS Feed.

    Parameters
    ----------
    get_result : pysnmp.smi.rfc1902.ObjectType
        Result from a pysnmp GET command.
    time : float
        Timestamp for when the SNMP GET was issued.

    Returns
    -------
    message : dict
        OCS Feed formatted message for publishing
    """
    message = {
        'block_name': 'syncbox',
        'timestamp': time,
        'data': {}
    }

    for item in get_result:
        field_name, oid_value, oid_description = _extract_oid_field_and_value(item)

        if oid_value is None:
            continue

        message['data'][field_name] = oid_value
        message['data'][field_name + "_description"] = oid_description

        if (field_name.split('_')[0] in CONVERT_OIDS):
            field_name, oid_value = _convert_oid_value(field_name, oid_value)
            message['data'][field_name] = oid_value
            message['data'][field_name + "_description"] = oid_description

    return message


def update_cache(get_result, timestamp):
    """Update the OID Value Cache.

    The OID Value Cache is used to store each unique OID and will be passed to
    session.data

    The cache consists of a dictionary, with the unique OIDs as keys, and
    another dictionary as the value. Each of these nested dictionaries contains
    the OID values, name, and description (decoded string). An example for a
    single OID, with connection status and timestamp information::

        {"mbgSyncboxN2XCurrentRefSource_0": {"status": 'PTP',
                                             "description": 'PTP'}}

    Additionally there is connection status and timestamp information under::

         "syncbox_connection": {"last_attempt": 1598543359.6326838,
                                "connected": True},
         "timestamp": 1656085022.680916}

    Parameters
    ----------
    get_result : pysnmp.smi.rfc1902.ObjectType
        Result from a pysnmp GET command.
    timestamp : float
        Timestamp for when the SNMP GET was issued.
    """
    oid_cache = {}
    # Return disconnected if SNMP response is empty
    if get_result is None:
        oid_cache['syncbox_connection'] = {'last_attempt': time.time(),
                                           'connected': False}
        return oid_cache

    for item in get_result:
        field_name, oid_value, oid_description = _extract_oid_field_and_value(item)
        if oid_value is None:
            continue

        # Update OID Cache for session.data
        oid_cache[field_name] = {"status": oid_value}
        oid_cache[field_name]["description"] = oid_description
        oid_cache['syncbox_connection'] = {'last_attempt': time.time(),
                                           'connected': True}
        oid_cache['timestamp'] = timestamp

        if (field_name.split('_')[0] in CONVERT_OIDS):
            field_name, oid_value = _convert_oid_value(field_name, oid_value)
            oid_cache[field_name] = {"status": oid_value}
            oid_cache[field_name]["description"] = oid_description

    return oid_cache


class MeinbergSyncboxAgent:
    """Monitor the syncbox system via SNMP.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    address : str
        Address of the syncbox.
    port : int
        SNMP port to issue GETs to, default to 161.
    version : int
        SNMP version for communication (1, 2, or 3), defaults to 1.
    outputs : list of ints
        List of outputs to monitor.

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    is_streaming : bool
        Tracks whether or not the agent is actively issuing SNMP GET commands
        to the syncbox. Setting to false stops sending commands.
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    """

    def __init__(self, agent, address, port=161, version=1, outputs=[1, 2, 3], sample_period=10):
        self.agent = agent
        self.is_streaming = False
        self.log = self.agent.log
        self.lock = TimeoutLock()

        self.log.info(f'Using SNMP version {version}.')
        self.version = version
        self.address = address
        self.snmp = SNMPTwister(address, port)
        self.connected = True

        self.lastGet = 0
        self.sample_period = sample_period

        # Create the list of OIDs to send get commands
        oids = ['mbgSyncboxN2XSerialNumber',
                'mbgSyncboxN2XFirmwareRevision',
                'mbgSyncboxN2XSystemTime',
                'mbgSyncboxN2XCurrentRefSource',
                'mbgSyncboxN2XPtpProfile',
                'mbgSyncboxN2XPtpNwProt',
                'mbgSyncboxN2XPtpPortState',
                'mbgSyncboxN2XPtpDelayMechanism',
                'mbgSyncboxN2XPtpDelayRequestInterval',
                'mbgSyncboxN2XPtpTimescale',
                'mbgSyncboxN2XPtpUTCOffset',
                'mbgSyncboxN2XPtpLeapSecondAnnounced',
                'mbgSyncboxN2XPtpGrandmasterClockID',
                'mbgSyncboxN2XPtpGrandmasterTimesource',
                'mbgSyncboxN2XPtpGrandmasterPriority1',
                'mbgSyncboxN2XPtpGrandmasterClockClass',
                'mbgSyncboxN2XPtpGrandmasterClockAccuracy',
                'mbgSyncboxN2XPtpGrandmasterClockVariance',
                'mbgSyncboxN2XPtpOffsetToGrandmaster',
                'mbgSyncboxN2XPtpMeanPathDelay']
        output_oids = ['mbgSyncboxN2XOutputMode']
        mib = 'MBG-SYNCBOX-N2X-MIB'

        self.get_list = []

        # Create the lists of OIDs to send get commands
        for oid in oids:
            self.get_list.append([(mib, oid, 0)])

        for out in outputs:
            for oid in output_oids:
                self.get_list.append([(mib, oid, out - 1)])

        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('syncbox',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('test_mode', default=False, type=bool)
    @inlineCallbacks
    def acq(self, session, params=None):
        """acq()

        **Process** - Fetch values from the syncbox via SNMP.

        Parameters
        ----------
        test_mode : bool, optional
            Run the Process loop only once. Meant only for testing.
            Default is False.

        Notes
        -----
        The most recent data collected is stored in session.data in the
        structure::

            >>> response.session['data']
            {'mbgSyncboxN2XSerialNumber_0':
                {'status': '009811006890',
                 'description': '009811006890'},
            'mbgSyncboxN2XFirmwareRevision_0':
                {'status': '1.20 ',
                 'description': '1.20 '},
            'mbgSyncboxN2XSystemTime_0':
                {'status': '2023-10-28 14:25:54 UTC',
                 'description': '2023-10-28 14:25:54 UTC'},
            'mbgSyncboxN2XCurrentRefSource_0':
                {'status': 'PTP',
                 'description': 'PTP'},
            'mbgSyncboxN2XPtpProfile_0':
                {'status': 0,
                 'description': 'none'},
            'mbgSyncboxN2XPtpNwProt_0':
                {'status': 1,
                 'description': 'ipv4'},
            'mbgSyncboxN2XPtpPortState_0':
                {'status': 9,
                 'description': 'slave'},
            'mbgSyncboxN2XPtpDelayMechanism_0':
                {'status': 0,
                 'description': 'e2e'},
            'mbgSyncboxN2XPtpDelayRequestInterval_0':
                {'status': 1,
                 'description': '1'},
            'mbgSyncboxN2XPtpTimescale_0':
                {'status': 0,
                 'description': 'tai'},
            'mbgSyncboxN2XPtpUTCOffset_0':
                {'status': '37 sec',
                 'description': '37 sec'},
            'mbgSyncboxN2XPtpUTCOffset_0_value':
                {'status': 37000000,
                 'description': '37 sec'},
            'mbgSyncboxN2XPtpLeapSecondAnnounced_0':
                {'status': 'no',
                 'description': 'no'},
            'mbgSyncboxN2XPtpGrandmasterClockID_0':
                {'status': 'EC:46:70:FF:FE:0A:AB:FE',
                 'description': 'EC:46:70:FF:FE:0A:AB:FE'},
            'mbgSyncboxN2XPtpGrandmasterTimesource_0':
                {'status': 32,
                 'description': 'gps'},
            'mbgSyncboxN2XPtpGrandmasterPriority1_0':
                {'status': 64,
                 'description': '64'},
            'mbgSyncboxN2XPtpGrandmasterClockClass_0':
                {'status': 6,
                 'description': '6'},
            'mbgSyncboxN2XPtpGrandmasterClockAccuracy_0':
                {'status': 33, 'description':
                 'accurateToWithin100ns'},
            'mbgSyncboxN2XPtpGrandmasterClockVariance_0':
                {'status': 13563,
                 'description': '13563'},
            'mbgSyncboxN2XPtpOffsetToGrandmaster_0':
                {'status': '10 ns',
                 'description': '10 ns'},
            'mbgSyncboxN2XPtpOffsetToGrandmaster_0_value':
                {'status': 0.01,
                 'description': '10 ns'},
            'mbgSyncboxN2XPtpMeanPathDelay_0':
                {'status': '875 ns',
                 'description': '875 ns'},
            'mbgSyncboxN2XPtpMeanPathDelay_0_value':
                {'status': 0.875,
                 'description': '875 ns'},
            'mbgSyncboxN2XOutputMode_1':
                {'status': 4,
                 'description': 'pulsePerSecond'},
             'syncbox_connection':
                {'last_attempt': 1656085022.680916,
                 'connected': True},
             'timestamp': 1656085022.680916,
             'address': '10.10.10.50'}

        Some relevant options and units for the above OIDs::

            mbgSyncboxN2XPtpProfile::
                Options:: none(0),
                          power(1),
                          telecom(2)
            mbgSyncboxN2XPtpNwProt::
                Options:: unknown(0),
                          ipv4(1),
                          ipv6(2),
                          ieee802-3(3),
                          deviceNet(4),
                          controlNet(5),
                          profiNet(6)
            mbgSyncboxN2XPtpPortState::
                Options:: uninitialized(0),
                          initializing(1),
                          faulty(2),
                          disabled(3),
                          listening(4),
                          preMaster(5),
                          master(6),
                          passive(7),
                          uncalibrated(8),
                          slave(9)
            mbgSyncboxN2XPtpDelayMechanism::
                Options:: e2e(0),
                          p2p(1)
            mbgSyncboxN2XPtpTimescale::
                Options:: tai(0),
                          arb(1)
            mbgSyncboxN2XPtpGrandmasterTimesource::
                Options:: atomicClock(16),
                          gps(32),
                          terrestrialRadio(48),
                          ptp(64),
                          ntp(80),
                          handSet(96),
                          other(144),
                          internalOscillator(160)
            mbgSyncboxN2XPtpGrandmasterClockAccuracy::
                Options:: accurateToWithin25ns(32),
                          accurateToWithin100ns(33),
                          accurateToWithin250ns(34),
                          accurateToWithin1us(35),
                          accurateToWithin2Point5us(36),
                          accurateToWithin10us(37),
                          accurateToWithin25us(38),
                          accurateToWithin100us(39),
                          accurateToWithin250us(40),
                          accurateToWithin1ms(41),
                          accurateToWithin2Point5ms(42),
                          accurateToWithin10ms(43),
                          accurateToWithin25ms(44),
                          accurateToWithin100ms(45),
                          accurateToWithin250ms(46),
                          accurateToWithin1s(47),
                          accurateToWithin10s(48),
                          accurateToGreaterThan10s(49)
            mbgSyncboxN2XPtpUTCOffset_value::
                Units:: microseconds
            mbgSyncboxN2XPtpOffsetToGrandmaster_value::
                Units:: microseconds
            mbgSyncboxN2XPtpMeanPathDelay_value::
                Units:: microseconds

        """

        self.is_streaming = True
        while self.is_streaming:
            yield dsleep(1)
            if not self.connected:
                self.log.error('No SNMP response. Check your connection!')
                self.log.info('Trying to reconnect.')

            read_time = time.time()

            # Check if sample period has passed before getting status
            if (read_time - self.lastGet) < self.sample_period:
                continue

            # Issue SNMP GET command
            # The syncbox has a unique case this requires issuing GET commands
            # one by one or else it will return the same data for each OID
            get_result = []
            for get in self.get_list:
                result = yield self.snmp.get(get, self.version)
                if result is None:
                    self.connected = False
                    session.data['syncbox_connection'] = {'last_attempt': time.time(),
                                                          'connected': False}
                    break
                get_result.extend(result)
                self.connected = True
            if not self.connected:
                session.degraded = True
                continue
            session.degraded = False

            # Do not publish if syncbox connection has dropped
            try:
                # Update session.data
                oid_cache = update_cache(get_result, read_time)
                oid_cache['address'] = self.address
                session.data = oid_cache
                self.log.info("{data}", data=session.data)

                self.lastGet = time.time()
                # Publish to feed
                message = _build_message(get_result, read_time)
                self.log.info("{msg}", msg=message)
                session.app.publish_to_feed('syncbox', message)
            except Exception as e:
                self.log.error(f'{e}')
                yield dsleep(1)

            if params['test_mode']:
                break

        return True, "Finished Recording"

    def _stop_acq(self, session, params=None):
        """_stop_acq()
        **Task** - Stop task associated with acq process.
        """
        if self.is_streaming:
            session.set_status('stopping')
            self.is_streaming = False
            return True, "Stopping Recording"
        else:
            return False, "Acq is not currently running"


def add_agent_args(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--address", help="Address to listen to.")
    pgroup.add_argument("--port", default=161,
                        help="Port to listen on.")
    pgroup.add_argument("--snmp-version", default='1', choices=['1', '2', '3'],
                        help="SNMP version for communication. Must match "
                             + "configuration on the syncbox.")
    pgroup.add_argument("--mode", default='acq', choices=['acq', 'test'])
    pgroup.add_argument("--outputs", nargs='+', default=[1, 2, 3], type=int,
                        help="Syncbox outputs to monitor. Defaults to [1,2,3].")
    pgroup.add_argument("--sample-period", default=10, type=int,
                        help="Acq sample period in seconds. Defaults to 10.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = add_agent_args()
    args = site_config.parse_args(agent_class='MeinbergSyncboxAgent',
                                  parser=parser,
                                  args=args)

    if args.mode == 'acq':
        init_params = True
    elif args.mode == 'test':
        init_params = False

    agent, runner = ocs_agent.init_site_agent(args)
    p = MeinbergSyncboxAgent(agent,
                             address=args.address,
                             port=int(args.port),
                             version=int(args.snmp_version),
                             outputs=args.outputs)

    agent.register_process("acq",
                           p.acq,
                           p._stop_acq,
                           startup=init_params, blocking=False)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



import struct

from pyModbusTCP.client import ModbusClient


class Hi6200Interface():
    """
    Connects to the Hi6200 weight sensor using a TCP ModbusClient with pyModbusTCP.
    The Modbus Client uses a socket connection to facillitate Modbus communications.
    ModbusClient requires an IP address and port to connect.

    The Gross and Net weight sensors are always available to read on the 8,9
    and 6,7 registers respectively.

    ModbusClient will not throw errors upon incorrect ip_address!

    ModbusClient auto-reconnects upon socket failure if auto_open is True.

    """

    def __init__(self, ip_address, tcp_port, verbose=False, **kwargs):
        self.scale = ModbusClient(host=ip_address,
                                  port=tcp_port,
                                  auto_open=True,
                                  auto_close=False)

    def _decode_scale_weight_registers(self, register_a, register_b):
        """
        Decodes the scales weight registers and returns a single weight value
        (float).

        The scale holds both the net and gross weights in permanent holding
        registers. Each weight is held across 2 registers in 4 hex bits (2 hex
        bits/4 bits per register, 8 bits total). The hex bits must be
        concatenated and converted to a float.

        """
        # Strip the '0x' hex bit
        # We must have 8 total bits to convert, so we zfill until each register
        # value is 4 bits
        hex_a = hex(register_b)[2:].zfill(4)
        hex_b = hex(register_b)[2:].zfill(4)

        # Concatenate the hex bits in cdab order.
        hex_weight = hex_b + hex_a

        # This struct function converts the concatenated hex bits to a float.
        return struct.unpack('!f', bytes.fromhex(hex_weight))[0]

    def read_scale_gross_weight(self):
        """
        Returns:
            float: The current gross weight reading of the scale in the sensors
            chosen unit (kg).

        """
        try:
            # The gross weight is always available on the 8,9 registers.
            # Reading these registers will return an int.
            a, b = self.scale.read_holding_registers(8, 2)

            return self._decode_scale_weight_registers(a, b)

        except AttributeError:
            return None

    def read_scale_net_weight(self):
        """
        Returns:
            float: The current net weight reading of the scale in the sensors
            chosen unit (kg).

        """
        try:
            # The gross weight is always available on the 6,7 registers.
            # Reading these registers will return an int.
            a, b = self.scale.read_holding_registers(6, 2)

            return self._decode_scale_weight_registers(a, b)

        except AttributeError:
            return None



import argparse
import time
from dataclasses import dataclass
from queue import Queue

import txaio
from twisted.internet import defer, reactor, threads

txaio.use_twisted()

from ocs import ocs_agent, site_config

import socs.agents.hwp_pmx.drivers.PMX_ethernet as pmx
from socs.agents.hwp_supervisor.agent import get_op_data


class Actions:
    class BaseAction:
        def __post_init__(self):
            self.deferred = defer.Deferred()
            self.log = txaio.make_logger()

    def process(self, *args, **kwargs):
        raise NotImplementedError

    @dataclass
    class SetOn(BaseAction):
        def process(self, module):
            self.log.info("Setting PMX on...")
            module.turn_on()

    @dataclass
    class SetOff(BaseAction):
        def process(self, module):
            self.log.info("Setting PMX off...")
            module.turn_off()

    @dataclass
    class UseExt(BaseAction):
        def process(self, module):
            self.log.info("Setting PMX Kikusui to PID control...")
            module.use_external_voltage()

    @dataclass
    class IgnExt(BaseAction):
        def process(self, module):
            self.log.info("Setting PMX Kikusui to direct control...")
            module.ign_external_voltage()

    @dataclass
    class ClearAlarm(BaseAction):
        def process(self, module):
            self.log.info("Clear PMX alarm...")
            module.clear_alarm()

    @dataclass
    class CancelShutdown(BaseAction):
        def process(self, module):
            self.log.info("Cancel shutdown...")

    @dataclass
    class CheckI(BaseAction):
        def process(self, module):
            msg, val = module.check_current()
            self.log.info(msg + "...")

    @dataclass
    class CheckV(BaseAction):
        def process(self, module):
            msg, val = module.check_voltage()
            self.log.info(msg + "...")

    @dataclass
    class SetI(BaseAction):
        curr: float

        def process(self, module):
            msg, val = module.set_current(self.curr)
            self.log.info(msg + "...")

    @dataclass
    class SetV(BaseAction):
        volt: float

        def process(self, module):
            msg, val = module.set_voltage(self.volt)
            self.log.info(msg + "...")

    @dataclass
    class CheckILim(BaseAction):
        def process(self, module):
            msg, val = module.check_current_limit()
            self.log.info(msg + "...")

    @dataclass
    class CheckVLim(BaseAction):
        def process(self, module):
            msg, val = module.check_voltage_limit()
            self.log.info(msg + "...")

    @dataclass
    class SetILim(BaseAction):
        curr: float

        def process(self, module):
            msg, val = module.set_current_limit(self.curr)
            self.log.info(msg + "...")

    @dataclass
    class SetVLim(BaseAction):
        volt: float

        def process(self, module):
            msg, val = module.set_voltage_limit(self.volt)
            self.log.info(msg + "...")


class HWPPMXAgent:
    """Agent for interfacing with a PMX Kikusui power supply
    to control the current and voltage that drive the rotation of the CHWP.

    Args:
        ip (str): IP address for the PMX Kikusui power supply
        port (str): Port for the PMX Kikusui power supply
        f_sample (float): sampling frequency (Hz)
        supervisor_id (str): Instance id of HWP supervisor
        no_data_timeout (float): Time (in seconds) to wait between receiving
            'no_data' actions from the supervisor and triggering a shutdown
    """

    def __init__(self, agent, ip, port, f_sample=1, supervisor_id=None,
                 no_data_timeout=15 * 60):

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log

        self.ip = ip
        self.port = port
        self.action_queue = Queue()

        self.f_sample = f_sample

        self._initialized = False
        self.take_data = False
        self.prot = 0

        self.shutdown_mode = False
        self.supervisor_id = supervisor_id
        self.no_data_timeout = no_data_timeout

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'hwppmx', record=True, agg_params=agg_params, buffer_time=1)

        self.agent.register_feed('rotation_action', record=True)

    @defer.inlineCallbacks
    def set_on(self, session, params):
        """set_on()
        **Task** - Turn on the PMX Kikusui.
        """
        action = Actions.SetOn(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set PMX Kikusui on'

    @defer.inlineCallbacks
    def set_off(self, session, params):
        """set_off()
        **Task** - Turn off the PMX Kikusui.
        """
        action = Actions.SetOff(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set PMX Kikusui off'

    @defer.inlineCallbacks
    def clear_alarm(self, session, params):
        """clear_alarm()
        **Task** - Clear alarm and exit protection mode.
        """
        action = Actions.ClearAlarm(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        self.prot = 0
        return True, 'Clear alarm'

    @defer.inlineCallbacks
    def use_ext(self, session, params):
        """use_ext()
        **Task** - Set the PMX Kikusui to use an external voltage control. Doing so
        enables PID control.
        """
        action = Actions.UseExt(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set PMX Kikusui to PID control'

    @defer.inlineCallbacks
    def ign_ext(self, session, params):
        """ign_ext()
        **Task** - Set the PMX Kiksui to ignore external voltage control. Doing so
        disables the PID and switches to direct control.
        """
        action = Actions.IgnExt(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set PMX Kikusui to direct control'

    @defer.inlineCallbacks
    def check_i(self, session, params):
        """check_i()
        **Task** - Set the current setting.
        """
        action = Actions.CheckI(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Check current is done'

    @defer.inlineCallbacks
    def check_v(self, session, params):
        """check_v()
        **Task** - Set the voltage setting.
        """
        action = Actions.CheckV(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Check voltage is done'

    @defer.inlineCallbacks
    @ocs_agent.param('curr', default=0, type=float)
    def set_i(self, session, params):
        """set_i(curr=0)
        **Task** - Set the current.

        Parameters:
            curr (float): set current
        """
        action = Actions.SetI(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set current is done'

    @defer.inlineCallbacks
    @ocs_agent.param('volt', default=0, type=float)
    def set_v(self, session, params):
        """set_v(volt=0)
        **Task** - Set the voltage.

        Parameters:
            volt (float): set voltage
        """
        action = Actions.SetV(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set voltage is done'

    @defer.inlineCallbacks
    def check_i_lim(self, session, params):
        """check_i_lim()
        **Task** - Check the current protection limit.
        """
        action = Actions.CheckILim(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Check current protection limit is done'

    @defer.inlineCallbacks
    def check_v_lim(self, session, params):
        """check_v_lim()
        **Task** - Check the voltage protection limit.
        """
        action = Actions.CheckVLim(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Check voltage protection limit is done'

    @defer.inlineCallbacks
    @ocs_agent.param('curr', default=1.3, type=float)
    def set_i_lim(self, session, params):
        """set_i_lim(curr=1.3)
        **Task** - Set the drive current limit.

        Parameters:
            curr (float): limit current
        """
        action = Actions.SetILim(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set voltage limit is done'

    @defer.inlineCallbacks
    @ocs_agent.param('volt', default=37., type=float)
    def set_v_lim(self, session, params):
        """set_v_lim(volt=37)
        **Task** - Set the drive voltage limit.

        Parameters:
            volt (float): limit voltage
        """
        action = Actions.SetVLim(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, 'Set current limit is done'

    @defer.inlineCallbacks
    def initiate_shutdown(self, session, params):
        """ initiate_shutdown()
        **Task** - Initiate the shutdown of the agent.
        """
        self.log.warn("INITIATING SHUTDOWN")

        action = Actions.SetOff(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        self.shutdown_mode = True
        return True, "Initiated shutdown mode"

    @defer.inlineCallbacks
    def cancel_shutdown(self, session, params):
        """cancel_shutdown()
        **Task** - Cancels shutdown mode, allowing other tasks to update the power supply
        """
        action = Actions.CancelShutdown(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        self.shutdown_mode = False
        return True, "Cancelled shutdown mode"

    def main(self, session, params):
        """main()

        **Process** - Start data acquisition.

        Notes:
            The most recent data collected is stored in session data in the
            structure::

                >>> response.session['data']
                {'curr': 0,
                 'volt': 0,
                 'prot': 0,
                 'last_updated': 1649085992.719602}

        """
        PMX = None

        threads.blockingCallFromThread(reactor, self._clear_queue)

        sleep_time = 1 / self.f_sample - 0.01
        last_daq = 0
        while session.status in ['starting', 'running']:
            if PMX is None:
                try:
                    PMX = pmx.PMX(ip=self.ip, port=self.port)
                except ConnectionRefusedError:
                    self.log.error(
                        "Could not connect to PMX. "
                        "Retrying after 30 sec..."
                    )
                    time.sleep(30)
                    continue

            now = time.time()
            if now - last_daq > sleep_time:
                self._get_and_publish_data(PMX, session)
                last_daq = now

            self._process_actions(PMX)
            time.sleep(0.1)

        PMX.close()
        return True, 'Stopped main'

    def _stop_main(self, session, params):
        """
        Stop acq process.
        """
        session.set_status('stopping')
        return True, 'Set main status to stopping'

    def _process_actions(self, PMX: pmx.PMX):
        while not self.action_queue.empty():
            action = self.action_queue.get()
            if action.__class__.__name__ in ['SetOn', 'SetOff', 'SetI', 'SetV', 'UseExt', 'IgnExt']:
                if self.shutdown_mode:
                    self.log.warn("Shutdown mode is in effect")
                    action.deferred.errback(Exception("Action cancelled by shutdown mode"))
                    return
            try:
                self.log.info(f"Running action {action}")
                res = action.process(PMX)
                reactor.callFromThread(action.deferred.callback, res)
            except Exception as e:
                self.log.error(f"Error processing action: {action}")
                reactor.callFromThread(action.deferred.errback, e)

    def _clear_queue(self):
        while not self.action_queue.empty():
            action = self.action_queue.get()
            action.deferred.errback(Exception("Action cancelled"))

    def _get_and_publish_data(self, PMX: pmx.PMX, session):
        now = time.time()
        data = {'timestamp': now,
                'block_name': 'hwppmx',
                'data': {}}

        try:
            msg, curr = PMX.meas_current()
            data['data']['current'] = curr

            msg, volt = PMX.meas_voltage()
            data['data']['voltage'] = volt

            msg, code = PMX.check_error()
            data['data']['err_code'] = code
            data['data']['err_msg'] = msg

            prot_code = PMX.check_prot()
            if prot_code != 0:
                self.prot = prot_code

            prot_msg = PMX.get_prot_msg(self.prot)
            data['data']['prot_code'] = self.prot
            data['data']['prot_msg'] = prot_msg

            msg, src = PMX.check_source()
            data['data']['source'] = src
            self.agent.publish_to_feed('hwppmx', data)
            session.data = {'curr': curr,
                            'volt': volt,
                            'prot': self.prot,
                            'prot_msg': prot_msg,
                            'source': src,
                            'last_updated': now}
        except BaseException:
            self.log.warn("Exception in getting data")
            return

    def monitor_supervisor(self, session, params):
        """monitor_supervisor()

        **Process** - This is a process that is constantly running to monitor the
        HWP supervisor and the recommended course of action
        course of action recommended by the HWP supervisor. If certain conditions
        are met, this will trigger a shutdown and force the power supply to
        power off.
        """

        last_ok_time = time.time()

        if self.supervisor_id is None:
            return False, "No supervisor ID set"

        while session.status in ['starting', 'running']:

            res = get_op_data(self.supervisor_id, 'monitor')
            if res['status'] != 'ok':
                action = 'no_data'
            else:
                action = res['data']['actions']['pmx']

            # If action is 'ok', update last_ok_time
            if action == 'ok':
                last_ok_time = time.time()

            # If action is 'no_data', check if last_ok_time, and potentially
            # trigger shutdown
            elif action == 'no_data':
                if (time.time() - last_ok_time) > self.no_data_timeout:
                    if not self.shutdown_mode:
                        self.agent.start('initiate_shutdown', params=None)

            # If action is 'shutdown', trigger shutdown
            elif action == 'stop':
                if not self.shutdown_mode:
                    self.agent.start('initiate_shutdown', params=None)

            data = {
                'data': {'rotation_action': action},
                'block_name': 'rotation_action',
                'timestamp': time.time()
            }

            self.agent.publish_to_feed('rotation_action', data)
            session.data = {
                'rotation_action': action,
                'time': time.time()
            }

            time.sleep(0.2)

        return True, 'Supervisor monitor has exited.'

    def _stop_monitor_supervisor(self, session, params):
        session.set_status('stopping')
        return True, "Stopping monitor shutdown."


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip', type=str,
                        help="ip address for kikusui PMX")
    pgroup.add_argument('--port', type=int,
                        help="port for kikusui PMX")
    pgroup.add_argument('--sampling-frequency', type=float, default=2.,
                        help="Sampling frequency for data acquisition")
    pgroup.add_argument('--supervisor-id', type=str,
                        help="Instance ID for HWP Supervisor agent")
    pgroup.add_argument('--no-data-timeout', type=float, default=15 * 60,
                        help="Time (sec) after which a 'no_data' action should "
                             "trigger a shutdown")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='HWPPMXAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    kwargs = {
        'ip': args.ip, 'port': args.port, 'supervisor_id': args.supervisor_id,
        'no_data_timeout': args.no_data_timeout
    }
    if args.sampling_frequency is not None:
        kwargs['f_sample'] = args.sampling_frequency
    PMX = HWPPMXAgent(agent, **kwargs)

    agent.register_process('main', PMX.main, PMX._stop_main, startup=True)
    agent.register_process(
        'monitor_supervisor', PMX.monitor_supervisor, PMX._stop_monitor_supervisor,
        startup=True)
    agent.register_task('set_on', PMX.set_on, blocking=False)
    agent.register_task('set_off', PMX.set_off, blocking=False)
    agent.register_task('clear_alarm', PMX.clear_alarm, blocking=False)
    agent.register_task('check_i', PMX.check_i, blocking=False)
    agent.register_task('check_v', PMX.check_v, blocking=False)
    agent.register_task('set_i', PMX.set_i, blocking=False)
    agent.register_task('set_v', PMX.set_v, blocking=False)
    agent.register_task('check_i_lim', PMX.check_i_lim, blocking=False)
    agent.register_task('check_v_lim', PMX.check_v_lim, blocking=False)
    agent.register_task('set_i_lim', PMX.set_i_lim, blocking=False)
    agent.register_task('set_v_lim', PMX.set_v_lim, blocking=False)
    agent.register_task('use_ext', PMX.use_ext, blocking=False)
    agent.register_task('ign_ext', PMX.ign_ext, blocking=False)
    agent.register_task('initiate_shutdown', PMX.initiate_shutdown, blocking=False)
    agent.register_task('cancel_shutdown', PMX.cancel_shutdown, blocking=False)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.wiregrid_tiltsensor.drivers import connect


class WiregridTiltSensorAgent:
    """ Agent to record the wiregrid tilt sensor data.
    The tilt sensor data is sent via serial-to-ethernet converter.

    Args:
        ip (str): IP address of the serial-to-ethernet converter
        port (int or str): Asigned port for the tilt sensor
            The converter has four D-sub ports to control
            multiple devices is determined
            by the ethernet port number of converter.
        sensor_type (str): Type of tilt sensor
            There are twp types of tilt sensor,
            and this argument is used for specifying
            to communicate with whichtilt sensor.
            This argument should be 'DWL' or 'sherborne'.
    """

    def __init__(self, agent, ip, port, sensor_type=None):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.take_data = False

        self.ip = ip
        self.port = int(port)
        self.sensor_type = sensor_type
        self.tiltsensor = connect(self.ip, self.port, self.sensor_type)
        self.pm = Pacemaker(2, quantize=True)

        agg_params = {'frame_length': 60}
        self.agent.register_feed('wgtiltsensor',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1.)

    def acq(self, session, params=None):
        """acq()

        **Process** - Run data acquisition.

        Notes:
            The most recent data collected is stored in session.data in the
            structure::

                >>> response.session['data']
                {'tiltsensor_data': {
                    'angleX': the angle in X-axis of tilt sensor,
                    'angleY': the angle in Y-axis of tilt sensor,
                    'temperatureX': the temperature in X-axis of tilt sensor
                                    this is available for only sherborne,
                    'temperatureY': the temperature in Y-axis of tilt sensor
                                    this is available for only sherborne
                    },
                'timestamp': timestamp when it updates tilt sensor data
                }
        """

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not start acq because {} is already running'
                )

            # Initialize a take_data flag
            self.take_data = True
            last_release = time.time()

            tiltsensor_data = {
                'timestamp': 0,
                'block_name': 'wgtiltsensor',
                'data': {
                    'angleX': -999,
                    'angleY': -999,
                    'temperatureX': -999,
                    'temperatureY': -999,
                },
            }

            self.log.info("Starting the count!")

            # Loop
            while self.take_data:
                # About every second, release and acquire the lock
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        print(f"Could not re-acquire lock now held by {self.lock.job}.")
                        return False

                # data taking
                current_time = time.time()
                msg, angles = self.tiltsensor.get_angle()
                if self.sensor_type == 'sherborne':
                    msg, temperatures = self.tiltsensor.get_temp()

                tiltsensor_data['timestamp'] = current_time
                tiltsensor_data['data']['angleX'] = angles[0]
                tiltsensor_data['data']['angleY'] = angles[1]
                if self.sensor_type == 'sherborne':
                    tiltsensor_data['data']['temperatureX'] = temperatures[0]
                    tiltsensor_data['data']['temperatureY'] = temperatures[1]

                self.agent.publish_to_feed('wgtiltsensor', tiltsensor_data)

                # store the session data
                session.data = {
                    'tiltsensor_data': {
                        'angleX': tiltsensor_data['data']['angleX'],
                        'angleY': tiltsensor_data['data']['angleY'],
                        'temperatureX': tiltsensor_data['data']['temperatureX'],
                        'temperatureY': tiltsensor_data['data']['temperatureY']
                    },
                    'timestamp': current_time
                }

                self.pm.sleep()  # DAQ interval
                # End of loop
            # End of lock acquiring

        self.agent.feeds['wgtiltsensor'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def stop_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop takeing data.'
        else:
            return False, 'acq is not currently running.'

    def reset(self, session, params=None):
        """reset()

        **Task** - Reset the tiltsensor if the type of tiltsensor is sherborne.

        Notes:
            The most recent data collected is stored in session.data in the
            structure::

                >>> response.session['data']
                {'reset': bool whether the reset successful or not
                 'timestamp': timestamp when this command is performed
                }
        """

        with self.lock.acquire_timeout(timeout=3.0, job='reset') as acquired:
            if not acquired:
                self.log.warn("Lock could not be acquired because it "
                              + f"is held by {self.lock.job}")
                return False

            if self.sensor_type != 'sherborne':
                return False, "This type of tiltsensor cannot reset."
            else:
                # Log the text provided to the Agent logs
                self.log.info("running reset")
                # Execute reset()
                self.tiltsensor.reset()
                # Store the timestamp when reset is performed in session.data
                session.data = {'reset': True,
                                'timestamp': time.time()}

        # True if task succeeds, False if not
        return True, 'Reset the tiltsensor'


def make_parser(parser_in=None):
    if parser_in is None:
        import argparse
        parser_in = argparse.ArgumentParser()

    pgroup = parser_in.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address', dest='ip', type=str, default=None,
                        help='The ip adress of the serial-to-ethernet converter')
    pgroup.add_argument('--port', dest='port', type=str, default=None,
                        help='The assigned port of the serial-to-ethernet converter '
                             'for the tilt sensor')
    pgroup.add_argument('--sensor-type',
                        dest='sensor_type',
                        type=str, default=None,
                        help='The type of tilt sensor '
                             'running wiregrid tilt sensor DAQ')
    return parser_in


def main(args=None):
    parser_in = make_parser()
    args = site_config.parse_args(agent_class='WiregridTiltSensorAgent',
                                  parser=parser_in,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    tiltsensor_agent = WiregridTiltSensorAgent(agent,
                                               ip=args.ip,
                                               port=args.port,
                                               sensor_type=args.sensor_type)

    agent.register_process('acq',
                           tiltsensor_agent.acq,
                           tiltsensor_agent.stop_acq,
                           startup=True)
    agent.register_task('reset', tiltsensor_agent.reset)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import time as tm

# Control modules
from socs.common import moxa_serial as mx


class Sherborne:
    """
    The sherborne object is for communicating with the sherborne tilt sensor

    Args:
    tcp_ip (str): TCP IP address
    tcp_port (int): TCP port
    """
    waittime = 0.05  # sec
    waittime_reset = 30  # sec, this is the time for the sensor to reset. should be over 30 sec
    address_Xaxis = 148  # default address of Xaxis is 148. this has individual sensor difference.
    address_Yaxis = 149  # default address of Yaxis is 149. this has individual sensor difference.

    command_angleX = b'!' + str(address_Xaxis).encode('ascii') + b':SYS?\r'
    command_angleY = b'!' + str(address_Yaxis).encode('ascii') + b':SYS?\r'
    command_resetX = b'!' + str(address_Xaxis).encode('ascii') + b':RST\r'
    command_resetY = b'!' + str(address_Yaxis).encode('ascii') + b':RST\r'

    def __init__(self, tcp_ip=None, tcp_port=None, timeout=None, reset_boot=False, verbose=0):
        self.tcp_ip = tcp_ip
        self.tcp_port = tcp_port
        self.verbose = verbose

        # Connect to device
        msg = self.__conn(tcp_ip, tcp_port, timeout)
        print(msg)
        if reset_boot:
            self.reset()

    def __del__(self):
        print(f"Disconnecting from TCP IP {self.tcp_ip} at port {self.tcp_port}")
        self.ser.close()
        return

    def get_angle(self):
        """ Measure the two-axis angle """
        self.clean_serial()
        if self.verbose > 0:
            print(f'get_angle() commands = {self.command_angleX}, {self.command_angleY}')
        SIZE = 16
        self.ser.write(self.command_angleX)
        read_angleX = self.ser.read(SIZE)
        value_read_angleX = read_angleX.decode('ascii')
        value_read_angleX = value_read_angleX.replace('\r', '')
        if self.verbose > 0:
            print(f'read_angleX = {value_read_angleX}')
        self.ser.write(self.command_angleY)
        read_angleY = self.ser.read(SIZE)
        value_read_angleY = read_angleY.decode('ascii')
        value_read_angleY = value_read_angleY.replace('\r', '')
        if self.verbose > 0:
            print(f'read_angleY = {value_read_angleY}')

        self.wait()

        val = (value_read_angleX, value_read_angleY)
        msg = f"Measured angle: X = {value_read_angleX}, Y = {value_read_angleY}"
        if self.verbose > 0:
            print(msg)

        return msg, val

    def __conn(self, tcp_ip, tcp_port, timeout):
        """
        Connect to the tilt sensor module

        Args:
        tcp_ip (str): TCP IP address
        tcp_port (int): TCP port
        """
        if tcp_ip is None or tcp_port is None:
            raise Exception(
                "Aborted Sherborne._conn() due to no TCP IP or "
                "TCP port specified")
        elif tcp_ip is not None and tcp_port is not None:
            self.ser = mx.Serial_TCPServer((tcp_ip, tcp_port), timeout, encoded=False)
            self.tcp_ip = tcp_ip
            self.tcp_port = int(tcp_port)
            self.using_tcp = True
            msg = f"Connected to TCP IP {tcp_ip} at port {tcp_port}"
        else:
            raise Exception(
                "Aborted Sherborne._conn() due to unknown error")
        return msg

    def wait(self):
        """ Sleep """
        tm.sleep(self.waittime)
        return True

    def wait_reset(self):
        """ Sleep for the reset time """
        tm.sleep(self.waittime_reset)
        return True

    def clean_serial(self):
        """ Flush the serial buffer """
        self.ser.flushInput()
        return True

    def reset(self):
        """ reset the tilt sensor """
        self.ser.write(self.command_resetX)
        readX = self.ser.read(2)
        if self.verbose > 0:
            print(readX)
        self.ser.write(self.command_resetY)
        readY = self.ser.read(2)
        if self.verbose > 0:
            print(readY)
        self.wait_reset()
        return True



# Sun Avoidance
#
# The docstring below is intended for injection into the documentation
# system.

"""When considering Sun Safety of a boresight pointing, we consider an
exclusion zone around the Sun with a user-specified radius. This is
called the ``exclusion_radius`` or the field-of-view radius.

The Safety of the instrument at any given moment is parametrized
by two numbers and a boolean:

  ``sun_dist``
    The separation between the Sun and the boresight (az, el), in
    degrees.  This is defined without regards for the horizon --
    i.e. even if the Sun and the boresight are both below "the
    horizon", the sun distance can be small.

  ``sun_time``
    The minimum time, in seconds, which must elapse before the current
    (az, el) pointing of the boresight will lie within the exclusion
    radius of the Sun.  Although some positions will "never" see the
    Sun, the sun_time for such positions is set to about 2 days, as a
    place-holder for "the future".  When the boresight is well below
    the horizon, the sun_time is set to "never", i.e. 2 days.  When
    the sun is below the horizon, then the sun_time will always be at
    least the time until next sun-rise.

  ``sun_down``
    A boolean indicating whether the sun is below the horizon
    elevation (which is a parameter of the instrument).


While the ``sun_dist`` and ``sun_down`` are important indicators of
whether the instrument is currently in immediate danger, the
``sun_time`` is useful for looking forward in order to avoid positions
that will soon be dangerous.

The user-defined policy for Sun Safety is captured in the following
settings:

  ``exclusion_radius``
    The radius, in degres, of a disk centered on the Sun that must be
    avoided by the boresight.

  ``min_sun_time``
    An (az, el) position is considered unsafe (danger zone) if the
    ``sun_time`` is less than the ``min_sun_time``. (Expressed in
    seconds.)

  ``response_time``
    An (az, el) position is considered vulnerable (warning zone) if
    the ``sun_time`` is less than the ``response_time``. This is
    intended to represent the maximum amount of time it could take an
    operator to reach the instrument and secure it, were motion to
    stall unexpectedly. (Expressed in seconds.)

  ``el_horizon``
    The elevation (in degrees) below which the Sun may be considered
    as invisible to the instrument.

  ``el_dodging``
    This setting affects how point-to-point motions are executed, with
    respect to what elevations may be used in intermediate legs of the
    trajectory. When this is False, the platform is restricted to
    travel only at elevations that lie between the initial and the
    target elevation. When True, the platform is permitted to travel
    at other elevations, all the way up to the limits of the
    platform. Using True is helpful to find Sun-safe trajectories in
    some circumstances. But False is helpful if excess elevation
    changes are potentially disturbing to the cryogenics.  This
    setting only affects point-to-point motions; "escape" paths will
    always consider all available elevations.

  ``axes_sequential``
    If True, a point-to-point motion will only be accepted if each leg
    is a constant el or constant az move. When this setting is False,
    legs that move simultaneously in az and el are permitted (and
    probably preferred) as long as they are safe.


A "Sun-safe" position is a pointing of the boresight that currently
has a ``sun_time`` that meets or exceeds the ``min_sun_time``
parameter.

"""

import datetime
import functools
import math
import time

import ephem
import numpy as np
from pixell import enmap
from so3g.proj import coords, quat

try:
    import pylab as pl
except ModuleNotFoundError:
    pass

DEG = np.pi / 180

HOUR = 3600
DAY = 86400
SIDEREAL_DAY = 86164.0905
NO_TIME = DAY * 2

#: Default policy to apply when evaluating Sun-safety and planning
#: trajectories.  Note the Agent code may apply different defaults,
#: based on known platform details.
DEFAULT_POLICY = {
    'exclusion_radius': 20,
    'min_el': 0,
    'max_el': 90,
    'min_az': -45,
    'max_az': 405,
    'el_horizon': 0,
    'el_dodging': False,
    'min_sun_time': HOUR,
    'response_time': HOUR * 4,
    'axes_sequential': False,
}


class SunTracker:
    """Provide guidance on what horizion coordinate positions and
    trajectories are sun-safe.

    Args:
      policy (dict): Exclusion policy parameters.  See module
        docstring, and DEFAULT_POLICY.  The policy should also include
        {min,max}\\_{el,az}, giving the limits supported by those axes.
      site (EarthlySite or None): Site to use; default is the SO LAT.
        If not None, pass an so3g.proj.EarthlySite or compatible.
      map_res (float, deg): resolution to use for the Sun Safety Map.
      sun_time_shift (float, seconds): For debugging and testing,
        compute the Sun's position as though it were this many
        seconds in the future.  If None or zero, this feature is
        disabled.
      fake_now (float, seconds): For debugging and testing, replace
        the tracker's computation of the current time (time.time())
        with this value.  If None, this testing feature is disabled.
      compute (bool): If True, immediately compute the Sun Safety Map
        by calling .reset().
      base_time (unix timestamp): Store this base_time and, if compute
        is True, pass it to .reset().

    """

    def __init__(self, policy=None, site=None,
                 map_res=.5, sun_time_shift=None, fake_now=None,
                 compute=True, base_time=None):
        # Note res is stored in radians.
        self.res = map_res * DEG
        if sun_time_shift is None:
            sun_time_shift = 0.
        self.sun_time_shift = sun_time_shift
        self.fake_now = fake_now
        self.base_time = base_time

        # Process and store the instrument config and safety policy.
        if policy is None:
            policy = {}
        for k in policy.keys():
            assert k in DEFAULT_POLICY
        _p = dict(DEFAULT_POLICY)
        _p.update(policy)
        self.policy = _p

        if site is None:
            # This is close enough.
            site = coords.SITES['so_lat']
        site_eph = ephem.Observer()
        site_eph.lon = site.lon * DEG
        site_eph.lat = site.lat * DEG
        site_eph.elevation = site.elev
        self._site = site_eph

        if compute:
            self.reset(base_time)

    def _now(self):
        if self.fake_now:
            return self.fake_now
        return time.time()

    def _sun(self, t):
        self._site.date = \
            datetime.datetime.utcfromtimestamp(t)
        return ephem.Sun(self._site)

    @staticmethod
    def _horizon_branch(az=0, el=0):
        """Given telescope boresight (el, az), which could be vectors,
        return (az_can, el_can, inverted), where el_can is in the
        branch most useful for Sun avoidance assessment, [-90, 90],
        and az_can is adjusted such that (az, el) and (az_can, el_can)
        correspond to same on-sky point.  For such points where the
        rebranching also implies an inversion of the focal plane,
        inverted is True.

        """
        q = quat.rotation_lonlat(-az * coords.DEG, el * coords.DEG)
        lon, lat, phi = quat.decompose_lonlat(q)
        inverted = np.zeros(lon.shape, bool)
        inverted[abs(phi) > .001 * coords.DEG] = True
        return -lon / coords.DEG, lat / coords.DEG, inverted

    def reset(self, base_time=None):
        """Compute and store the Sun Safety Map for a specific
        timestamp.

        This basic computation is required prior to calling other
        functions that use the Sun Safety Map.

        """
        # Set a reference time -- the map of sun times is usable from
        # this reference time to at least 12 hours in the future.
        if base_time is None:
            base_time = self._now()

        # Map extends from dec -80 to +80.
        shape, wcs = enmap.band_geometry(
            dec_cut=80 * DEG, res=self.res, proj='car')

        # Map where each pixel is distance to the Sun.
        sun_dist = enmap.zeros(shape, wcs=wcs) - 1

        # Map of time-to-sun-danger, when sun is above horizon.
        sun_times_up = sun_dist.copy()

        # Map of time-to-sun-danger, when sun is below horizon.
        sun_times_dn = sun_dist.copy()

        # Quaternion rotation for each point in the map.
        dec, ra = sun_dist.posmap()
        map_q = quat.rotation_lonlat(ra.ravel(), dec.ravel())

        # Look up the sun position at our shifted time; but also shift
        # explicitly in RA for earth rotation.
        v = self._sun(base_time + self.sun_time_shift)
        sun_dec = v.dec
        sun_ra = v.ra - self.sun_time_shift * (2 * np.pi / SIDEREAL_DAY)
        del v

        # Compute the map of angular distance to the Sun, at
        # base_time.
        qsun = quat.rotation_lonlat(sun_ra, sun_dec)
        qoff = ~qsun * map_q
        sun_dist[:] = (quat.decompose_iso(~qoff)[0]
                       .reshape(sun_dist.shape) / DEG)

        # For the sun_times_* maps, each pixel will give the time
        # delay between base_time and the time when the sky coordinate
        # will be inside the Sun exclusion mask.
        dt = -ra[0] * DAY / (2 * np.pi)

        # For sun_times_up (sun above horizon), the region around the
        # sun is bad right now.
        sun_times_up[sun_dist <= self.policy['exclusion_radius']] = 0.

        # Loop over rows of the sun_times_* maps.
        for g_up, g_dn in zip(sun_times_up, sun_times_dn):
            if (g_up < 0).all():
                # Sun mask does not touch this declination.
                continue
            # Identify pixel on the right of the masked region.
            flips = ((g_up == 0)
                     * np.hstack((g_up[:-1] != g_up[1:],
                                  g_up[-1] != g_up[0]))).nonzero()[0]
            dt0 = dt[flips[0]]
            _dt = (dt - dt0) % DAY
            # For sun_times_up, fill only pixels outside sun mask.
            g_up[g_up < 0] = _dt[g_up < 0]
            # For sun_times_dn, fill all pixels.
            g_dn[:] = _dt[:]

        # Fill in remaining -1 with NO_TIME.
        sun_times_up[sun_times_up < 0] = NO_TIME
        sun_times_dn[sun_times_dn < 0] = NO_TIME

        # Store the sun_times map and stuff.
        self.base_time = base_time
        self.sun_times_up = sun_times_up
        self.sun_times_dn = sun_times_dn
        self.sun_dist = sun_dist
        self.map_q = map_q

    def _azel_pix(self, az, el, dt=0, round=True, segments=False):
        """Return the pixel indices of the Sun Safety Map that are
        hit by the trajectory (az, el) at time dt.

        Args:
          az (array of float, deg): Azimuth.
          el (array of float, deg): Elevation.
          dt (array of float, s): Time offset relative to the base
            time, at which to evaluate the trajectory.
          round (bool): If True, round results to integer (for easy
            look-up in the map).
          segments (bool): If True, split up the trajectory into
            segments (a list of pix_ji sections) such that they don't
            cross the map boundaries at any point.

        """
        az = np.asarray(az)
        el = np.asarray(el)
        qt = coords.CelestialSightLine.naive_az_el(
            self.base_time + dt, az * DEG, el * DEG).Q
        ra, dec, _ = quat.decompose_lonlat(qt)
        src_map = self.sun_dist  # Only used for coordinate ops.
        pix_ji = src_map.sky2pix((dec, ra))
        if round:
            pix_ji = pix_ji.round().astype(int)
            # Handle out of bounds as follows:
            # - RA indices are mod-ed into range.
            # - dec indices are clamped to the map edge.
            j, i = pix_ji
            j[j < 0] = 0
            j[j >= src_map.shape[-2]] = src_map.shape[-2] - 1
            i[:] = i % src_map.shape[-1]

        if segments:
            jumps = ((abs(np.diff(pix_ji[0])) > src_map.shape[-2] / 2)
                     + (abs(np.diff(pix_ji[1])) > src_map.shape[-1] / 2))
            jump = jumps.nonzero()[0]
            starts = np.hstack((0, jump + 1))
            stops = np.hstack((jump + 1, len(pix_ji[0])))
            return [pix_ji[:, a:b] for a, b in zip(starts, stops)]

        return pix_ji

    def check_trajectory(self, az, el, t=None, raw=False):
        """For a telescope trajectory (vectors az, el, in deg), assumed to
        occur at time t (defaults to now), get the minimum value of
        the Sun Safety Map traversed by that trajectory.  Also get the
        minimum value of the Sun Distance map.

        This requires the Sun Safety Map to have been computed with a
        base_time in the 24 hours before t.

        Returns a dict with entries:

        - ``'sun_time'``: Minimum Sun Safety Time on the traj.
        - ``'sun_time_start'``: Sun Safety Time at first point.
        - ``'sun_time_stop'``: Sun Safety Time at last point.
        - ``'sun_dist_min'``: Minimum distance to Sun, in degrees.
        - ``'sun_dist_mean'``: Mean distance to Sun.
        - ``'sun_dist_start'``: Distance to Sun, at first point.
        - ``'sun_dist_stop'``: Distance to Sun, at last point.

        """
        if t is None:
            t = self._now()
        az, el = np.asarray(az), np.asarray(el)
        j, i = self._azel_pix(az, el, dt=t - self.base_time)
        sun_delta = self.sun_times_up[j, i]
        sun_dists = self.sun_dist[j, i]

        # If the sun is below the horizon, sun times are modified.
        az_sun, el_sun = self.get_sun_pos(t=t)['sun_azel']
        if el_sun < self.policy['el_horizon']:
            if (az_sun % 360.) > 180:
                # The setting problem:
                sun_delta = self.sun_times_dn[j, i]
            else:
                # The rising problem:
                dt_rise = self._next_rise_time(t) - t
                sun_delta[sun_delta < dt_rise] = dt_rise

        # Positions below the modified horizon are always safe.
        safe_el = self.policy['el_horizon'] - self.policy['exclusion_radius']
        _, el_can, _ = self._horizon_branch(az, el)
        sun_delta[el_can < safe_el] = NO_TIME

        if raw:
            return sun_delta, sun_dists
        return {
            'sun_time': sun_delta.min(),
            'sun_time_start': sun_delta[0],
            'sun_time_stop': sun_delta[-1],
            'sun_dist_start': sun_dists[0],
            'sun_dist_stop': sun_dists[-1],
            'sun_dist_min': sun_dists.min(),
            'sun_dist_mean': sun_dists.mean(),
        }

    def get_sun_pos(self, az=None, el=None, t=None):
        """Get info on the Sun's location at time t.  If (az, el) are also
        specified, returns the angular separation between that
        pointing and Sun's center.

        """
        if t is None:
            t = self._now()
        eff_t = t + self.sun_time_shift

        v = self._sun(eff_t)
        qsun = quat.rotation_lonlat(v.ra, v.dec)
        qzen = coords.CelestialSightLine.naive_az_el(eff_t, 0, np.pi / 2).Q
        neg_sun_az, sun_el, _ = quat.decompose_lonlat(~qzen * qsun)

        results = {
            'sun_radec': (v.ra / DEG, v.dec / DEG),
            'sun_azel': ((-neg_sun_az / DEG) % 360., sun_el / DEG),
            'sun_down': sun_el / DEG < self.policy['el_horizon'],
        }
        if self.sun_time_shift != 0:
            results['WARNING'] = 'Fake Sun Position is in use!'

        if az is not None:
            qtel = coords.CelestialSightLine.naive_az_el(
                eff_t, az * DEG, el * DEG).Q
            r = quat.decompose_iso(~qtel * qsun)[0]
            results['sun_dist'] = r / DEG
            _, el_can, _ = self._horizon_branch(az, el)
            results['platform_down'] = \
                el_can < (self.policy['el_horizon'] - self.policy['exclusion_radius'])
        return results

    def _next_rise_time(self, t):
        """Compute the smallest time, greater than or equal to t, at
        which the Sun will be above the el_horizon.  Accurate to 1s or
        so.

        """
        # Since the helper is cached, include anything in the args
        # that could affect the computed value.
        return self._next_rise_time_cache_helper(t, self.sun_time_shift)

    @functools.lru_cache
    def _next_rise_time_cache_helper(self, t, *args):
        horizon = self.policy['el_horizon']
        el0 = self.get_sun_pos(t=t)['sun_azel'][1]
        if el0 > horizon:
            return t
        step = 3600
        while abs(step) > 1.:
            t += step
            el = self.get_sun_pos(t=t)['sun_azel'][1]
            if el > horizon:
                step = -abs(step) / 2
            else:
                step = abs(step)
        return t

    def show_map(self, axes=None, show=True):
        """Plot the Sun Safety Map and Sun Distance Map on the provided axes
        (a list)."""
        if axes is None:
            fig, axes = pl.subplots(2, 1)
            fig.tight_layout()
        else:
            fig = None

        imgs = []
        for axi, ax in enumerate(axes):
            if axi == 0:
                # Sun safe time
                x = self.sun_times_up / HOUR
                x[x == NO_TIME] = np.nan
                title = 'Sun safe time (hours)'
            elif axi == 1:
                # Sun distance
                x = self.sun_dist
                title = 'Sun distance (degrees)'
            im = ax.imshow(x, origin='lower', cmap='Oranges')
            ji = self._azel_pix(0, np.array([90.]))
            ax.scatter(ji[1], ji[0], marker='x', color='white')
            ax.set_title(title)
            pl.colorbar(im, ax=ax)
            imgs.append(im)

        if show:
            pl.show()

        return fig, axes, imgs

    def analyze_paths(self, az0, el0, az1, el1, t=None,
                      plot_file=None, dodging=True):
        """Design and analyze a number of different paths between (az0, el0)
        and (az1, el1).  Return the list, for further processing and
        choice.

        """
        if t is None:
            t = self._now()

        if plot_file:
            assert (t == self.base_time)  # Can only plot "now" results.
            fig, axes, imgs = self.show_map(show=False)
            last_el = None

        # Test all trajectories with intermediate el.
        all_moves = []

        base = {
            'req_start': (az0, el0),
            'req_stop': (az1, el1),
            'req_time': t,
            'travel_el': (el0 + el1) / 2,
            'travel_el_confined': True,
            'direct': True,
        }

        # Suitable list of test els.
        el_lims = [self.policy[_k] for _k in ['min_el', 'max_el']]
        if el0 == el1:
            el_nodes = [el0]
        else:
            el_nodes = sorted([el0, el1])
        if dodging and (el_lims[0] < el_nodes[0]):
            el_nodes.insert(0, el_lims[0])
        if dodging and (el_lims[1] > el_nodes[-1]):
            el_nodes.append(el_lims[1])

        el_sep = 1.
        el_cands = []
        for i in range(len(el_nodes) - 1):
            n = math.ceil((el_nodes[i + 1] - el_nodes[i]) / el_sep)
            assert (n >= 1)
            el_cands.extend(list(
                np.linspace(el_nodes[i], el_nodes[i + 1], n + 1)[:-1]))
        el_cands.append(el_nodes[-1])

        for iel in el_cands:
            detail = dict(base)
            detail.update({
                'direct': False,
                'travel_el': iel,
                'travel_el_confined': (iel >= min(el0, el1)) and (iel <= max(el0, el1)),
            })
            moves = MoveSequence(az0, el0, az0, iel, az1, iel, az1, el1, simplify=True)

            detail['moves'] = moves
            traj_info = self.check_trajectory(*moves.get_traj(), t=t)
            detail.update(traj_info)
            all_moves.append(detail)
            if plot_file and (last_el is None or abs(last_el - iel) > 5):
                c = 'black'
                for j, i in self._azel_pix(*moves.get_traj(), round=True, segments=True):
                    for ax in axes:
                        a, = ax.plot(i, j, color=c, lw=1)
                last_el = iel

        # Include the direct path, but put in "worst case" details
        # based on all "confined" paths computed above.
        direct = dict(base)
        direct['moves'] = MoveSequence(az0, el0, az1, el1, simplify=True)
        traj_info = self.check_trajectory(*direct['moves'].get_traj(), t=t)
        direct.update(traj_info)
        conf = [m for m in all_moves if m['travel_el_confined']]
        if len(conf):
            for k in ['sun_time', 'sun_dist_min', 'sun_dist_mean']:
                direct[k] = min([m[k] for m in conf])
            all_moves.append(direct)

        if plot_file:
            # Add the direct traj, in blue.
            segments = self._azel_pix(*direct['moves'].get_traj(), round=True, segments=True)
            for ax in axes:
                for j, i in segments:
                    ax.plot(i, j, color='blue')
                for seg, rng, mrk in [(segments[0], slice(0, 1), 'o'),
                                      (segments[-1], slice(-1, None), 'x')]:
                    ax.scatter(seg[1][rng], seg[0][rng], marker=mrk, color='blue')
            # Add the selected trajectory in green.
            selected = self.select_move(all_moves)[0]
            if selected is not None:
                traj = selected['moves'].get_traj()
                segments = self._azel_pix(*traj, round=True, segments=True)
                for ax in axes:
                    for j, i in segments:
                        ax.plot(i, j, color='green')

            pl.savefig(plot_file)
        return all_moves

    def find_escape_paths(self, az0, el0, t=None,
                          debug=False):
        """Design and analyze a number of different paths that move from (az0,
        el0) to a sun safe position.  Return the list, for further
        processing and choice.

        """
        if t is None:
            t = self._now()

        az_cands = []
        _az = math.ceil(self.policy['min_az'] / 180) * 180
        while _az <= self.policy['max_az']:
            az_cands.append(_az)
            _az += 180.

        # Clip el0 into the allowed range.
        el0 = np.clip(el0, self.policy['min_el'], self.policy['max_el'])

        # Preference is to not change altitude; but allow for lowering
        # (in absolute terms) -- there is "more sky" when you move
        # away from zenith.
        if el0 <= 90:
            n_el = math.ceil(el0 - self.policy['min_el']) + 1
            els = np.linspace(el0, self.policy['min_el'], n_el)
        else:
            n_el = math.ceil(self.policy['max_el'] - el0) + 1
            els = np.linspace(el0, self.policy['max_el'], n_el)

        path = None
        for el1 in els:
            paths = [self.analyze_paths(az0, el0, _az, el1, t=t, dodging=False)
                     for _az in az_cands]
            best_paths = [self.select_move(p)[0] for p in paths]
            best_paths = [p for p in best_paths if p is not None]
            if len(best_paths):
                path = self.select_move(best_paths)[0]
                if debug:
                    cands, _ = self.select_move(best_paths, raw=True)
                    return cands
            if path is not None:
                return path

        return None

    def select_move(self, moves, raw=False):
        """Given a list of possible "moves", select the best one.
        The "moves" should be like the ones returned by
        ``analyze_paths``.

        The best move is determined by first screening out dangerous
        paths (ones that pass close to Sun, move closer to Sun
        unnecessarily, violate axis limits, etc.) and then identifying
        paths that minimize danger (distance to Sun; Sun time) and
        path length.

        If raw=True, a debugging output is returned; see code.

        Returns:
          (dict, list): (best_move, decisions)

          ``best_move`` -- the element of moves that is safest.  If no
          safe move was found, None is returned.

          ``decisions`` - List of dicts, in one-to-one correspondence
          with ``moves``.  Each decision dict has entries 'rejected'
          (True or False) and 'reason' (string description of why the
          move was rejected outright).

        """
        _p = self.policy

        decisions = [{'rejected': False,
                      'reason': None} for m in moves]

        def reject(d, reason):
            d['rejected'] = True
            d['reason'] = reason

        # According to policy, reject moves outright.
        for m, d in zip(moves, decisions):
            if d['rejected']:
                continue

            els = m['req_start'][1], m['req_stop'][1]

            if _p['axes_sequential'] and m['moves'].includes_mixed_moves():
                reject(d, 'Path includes simultaneous az+el legs (forbidden in policy).')
                continue

            if (m['sun_time_start'] < _p['min_sun_time']):
                # If the path is starting in danger zone, then only
                # enforce that the move takes the platform to a better place.

                # Test > res, rather than > 0... near the minimum this
                # can be noisy.
                if m['sun_dist_start'] - m['sun_dist_min'] > self.res / DEG:
                    reject(d, 'Path moves even closer to sun.')
                    continue
                if m['sun_time_stop'] < _p['min_sun_time']:
                    reject(d, 'Path does not end in sun-safe location.')
                    continue

            elif m['sun_time'] < _p['min_sun_time']:
                reject(d, 'Path too close to sun.')
                continue

            if m['travel_el'] < _p['min_el']:
                reject(d, 'Path goes below minimum el.')
                continue

            if m['travel_el'] > _p['max_el']:
                reject(d, 'Path goes above maximum el.')
                continue

            if not _p['el_dodging']:
                if m['travel_el'] < min(*els):
                    reject(d, 'Path dodges (goes below necessary el range).')
                    continue
                if m['travel_el'] > max(*els):
                    reject(d, 'Path dodges (goes above necessary el range).')

        cands = [m for m, d in zip(moves, decisions)
                 if not d['rejected']]
        if len(cands) == 0:
            return None, decisions

        def metric_func(m):
            # Sorting key for move proposals.  More preferable paths
            # should have higher sort order.
            azs = m['req_start'][0], m['req_stop'][0]
            els = m['req_start'][1], m['req_stop'][1]
            return (
                # Low sun_time is bad, though anything longer
                # than response_time is equivalent.
                m['sun_time'] if m['sun_time'] < _p['response_time'] else _p['response_time'],

                # Single leg moves are preferred, for simplicity.
                m['direct'],

                # Higher minimum sun distance is preferred.
                m['sun_dist_min'],

                # Shorter paths (less total az / el motion) are preferred.
                -(abs(m['travel_el'] - els[0]) + abs(m['travel_el'] - els[1])),
                -abs(azs[1] - azs[0]),

                # Larger mean Sun distance is preferred.  But this is
                # subdominant to path length; otherwise spinning
                # around a bunch of times can be used to lower the
                # mean sun dist!
                m['sun_dist_mean'],

                # Prefer higher elevations for the move, all else being equal.
                m['travel_el'],
            )
        cands.sort(key=metric_func)
        if raw:
            return [(c, metric_func(c)) for c in cands], decisions
        return cands[-1], decisions


class MoveSequence:
    def __init__(self, *args, simplify=False):
        """Container for a series of (az, el) positions.  Pass the
        positions to the constructor as (az, el) tuples::

          MoveSequence((60, 180), (60, 90), (50, 90))

        or equivalently as individual arguments::

          MoveSequence(60, 180, 60, 90, 50, 90)

        If simplify=True is passed, then any immediate position
        repetitions are deleted.

        """
        self.nodes = []
        if len(args) == 0:
            return
        is_tuples = [isinstance(a, tuple) for a in args]
        if all(is_tuples):
            pass
        elif any(is_tuples):
            raise ValueError('Constructor accepts tuples or az, el, az, el; not a mix.')
        else:
            assert (len(args) % 2 == 0)
            args = [(args[i], args[i + 1]) for i in range(0, len(args), 2)]
        for (az, el) in args:
            self.nodes.append((float(az), float(el)))
        if simplify:
            # Remove repeated nodes.
            idx = 0
            while idx < len(self.nodes) - 1:
                if self.nodes[idx] == self.nodes[idx + 1]:
                    self.nodes.pop(idx + 1)
                else:
                    idx += 1

    def get_legs(self):
        """Iterate over the legs of the MoveSequence; yields each ((az_start,
        el_start), (az_end, az_end)).

        """
        for i in range(len(self.nodes) - 1):
            yield self.nodes[i:i + 2]

    def get_traj(self, res=0.5):
        """Return (az, el) vectors with the full path for the MoveSequence.
        No step in az or el will be greater than res.

        """
        if len(self.nodes) == 1:
            return np.array([self.nodes[0][0]]), np.array([self.nodes[0][1]])

        xx, yy = [], []
        for (x0, y0), (x1, y1) in self.get_legs():
            n = max(2, math.ceil(abs(x1 - x0) / res), math.ceil(abs(y1 - y0) / res))
            xx.append(np.linspace(x0, x1, n))
            yy.append(np.linspace(y0, y1, n))
        return np.hstack(tuple(xx)), np.hstack(tuple(yy))

    def includes_mixed_moves(self):
        """Returns True if any of the legs include simultaneous
        changes in az and el.

        """
        for (x0, y0), (x1, y1) in self.get_legs():
            if (x0 != x1) and (y0 != y1):
                return True
        return False



# SNMP MIB module (IBOOTBAR-MIB) expressed in pysnmp data model.
#
# This Python module is designed to be imported and executed by the
# pysnmp library.
#
# See https://www.pysnmp.com/pysnmp for further information.
#
# Notes
# -----
# ASN.1 source file://./IBOOTBAR-MIB.mib
# Produced by pysmi-1.5.9 at Tue Apr 22 14:30:23 2025
# On host login platform Linux version 5.15.0-136-generic by user ykyohei
# Using Python version 3.10.12 (main, Jul  5 2023, 18:54:27) [GCC 11.2.0]

if 'mibBuilder' not in globals():
    import sys

    sys.stderr.write(__doc__)
    sys.exit(1)

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.importSymbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.importSymbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.importSymbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(ModuleCompliance,
 NotificationGroup) = mibBuilder.importSymbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 enterprises,
 iso) = mibBuilder.importSymbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "enterprises",
    "iso")

(DisplayString,
 PhysAddress,
 TextualConvention) = mibBuilder.importSymbols(
    "SNMPv2-TC",
    "DisplayString",
    "PhysAddress",
    "TextualConvention")


# MODULE-IDENTITY

dataprobe = ModuleIdentity(
    (1, 3, 6, 1, 4, 1, 1418)
)


# Types definitions


# TEXTUAL-CONVENTIONS



class TC1(TextualConvention, Integer32):
    status = "current"


# MIB Managed Objects in the order of their OIDs

_IBootBarAgent_ObjectIdentity = ObjectIdentity
iBootBarAgent = _IBootBarAgent_ObjectIdentity(
    (1, 3, 6, 1, 4, 1, 1418, 4)
)
_SystemSettings_ObjectIdentity = ObjectIdentity
systemSettings = _SystemSettings_ObjectIdentity(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1)
)
_DeviceName_Type = DisplayString
_DeviceName_Object = MibScalar
deviceName = _DeviceName_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 1),
    _DeviceName_Type()
)
deviceName.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    deviceName.setStatus("current")


class _IpMode_Type(Integer32):
    """Custom type ipMode based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2)
        )
    )
    namedValues = NamedValues(
        *(("static", 0),
          ("arp-ping", 1),
          ("dhcp", 2))
    )


_IpMode_Type.__name__ = "Integer32"
_IpMode_Object = MibScalar
ipMode = _IpMode_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 2),
    _IpMode_Type()
)
ipMode.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    ipMode.setStatus("current")


class _IpAddress_Type(DisplayString):
    """Custom type ipAddress based on DisplayString"""
    subtypeSpec = DisplayString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(16, 16),
    )
    fixedLength = 16


_IpAddress_Type.__name__ = "DisplayString"
_IpAddress_Object = MibScalar
ipAddress = _IpAddress_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 3),
    _IpAddress_Type()
)
ipAddress.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    ipAddress.setStatus("current")


class _SubnetMask_Type(DisplayString):
    """Custom type subnetMask based on DisplayString"""
    subtypeSpec = DisplayString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(16, 16),
    )
    fixedLength = 16


_SubnetMask_Type.__name__ = "DisplayString"
_SubnetMask_Object = MibScalar
subnetMask = _SubnetMask_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 4),
    _SubnetMask_Type()
)
subnetMask.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    subnetMask.setStatus("current")


class _Gateway_Type(DisplayString):
    """Custom type gateway based on DisplayString"""
    subtypeSpec = DisplayString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(16, 16),
    )
    fixedLength = 16


_Gateway_Type.__name__ = "DisplayString"
_Gateway_Object = MibScalar
gateway = _Gateway_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 5),
    _Gateway_Type()
)
gateway.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    gateway.setStatus("current")


class _WebEnable_Type(Integer32):
    """Custom type webEnable based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("false", 0),
          ("true", 1))
    )


_WebEnable_Type.__name__ = "Integer32"
_WebEnable_Object = MibScalar
webEnable = _WebEnable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 6),
    _WebEnable_Type()
)
webEnable.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    webEnable.setStatus("current")


class _WebPort_Type(Integer32):
    """Custom type webPort based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 65535),
    )


_WebPort_Type.__name__ = "Integer32"
_WebPort_Object = MibScalar
webPort = _WebPort_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 7),
    _WebPort_Type()
)
webPort.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    webPort.setStatus("current")


class _SslEnable_Type(Integer32):
    """Custom type sslEnable based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("false", 0),
          ("true", 1))
    )


_SslEnable_Type.__name__ = "Integer32"
_SslEnable_Object = MibScalar
sslEnable = _SslEnable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 8),
    _SslEnable_Type()
)
sslEnable.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    sslEnable.setStatus("current")


class _TelnetEnable_Type(Integer32):
    """Custom type telnetEnable based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("false", 0),
          ("true", 1))
    )


_TelnetEnable_Type.__name__ = "Integer32"
_TelnetEnable_Object = MibScalar
telnetEnable = _TelnetEnable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 9),
    _TelnetEnable_Type()
)
telnetEnable.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    telnetEnable.setStatus("current")


class _TelnetPort_Type(Integer32):
    """Custom type telnetPort based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 65535),
    )


_TelnetPort_Type.__name__ = "Integer32"
_TelnetPort_Object = MibScalar
telnetPort = _TelnetPort_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 10),
    _TelnetPort_Type()
)
telnetPort.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    telnetPort.setStatus("current")


class _UpdateEnable_Type(Integer32):
    """Custom type updateEnable based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("false", 0),
          ("true", 1))
    )


_UpdateEnable_Type.__name__ = "Integer32"
_UpdateEnable_Object = MibScalar
updateEnable = _UpdateEnable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 11),
    _UpdateEnable_Type()
)
updateEnable.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    updateEnable.setStatus("current")
_CycleTime_Type = Integer32
_CycleTime_Object = MibScalar
cycleTime = _CycleTime_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 12),
    _CycleTime_Type()
)
cycleTime.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    cycleTime.setStatus("current")
_DelayTime_Type = Integer32
_DelayTime_Object = MibScalar
delayTime = _DelayTime_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 1, 13),
    _DelayTime_Type()
)
delayTime.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    delayTime.setStatus("current")
_SnmpManagerTable_Object = MibTable
snmpManagerTable = _SnmpManagerTable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 2)
)
if mibBuilder.loadTexts:
    snmpManagerTable.setStatus("current")
_SnmpManagerEntry_Object = MibTableRow
snmpManagerEntry = _SnmpManagerEntry_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 2, 1)
)
snmpManagerEntry.setIndexNames(
    (0, "IBOOTBAR-MIB", "snmpManagerIndex"),
)
if mibBuilder.loadTexts:
    snmpManagerEntry.setStatus("current")
_SnmpManagerIndex_Type = Integer32
_SnmpManagerIndex_Object = MibTableColumn
snmpManagerIndex = _SnmpManagerIndex_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 2, 1, 1),
    _SnmpManagerIndex_Type()
)
snmpManagerIndex.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    snmpManagerIndex.setStatus("current")


class _SnmpManagerIPAddress_Type(DisplayString):
    """Custom type snmpManagerIPAddress based on DisplayString"""
    subtypeSpec = DisplayString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(16, 16),
    )
    fixedLength = 16


_SnmpManagerIPAddress_Type.__name__ = "DisplayString"
_SnmpManagerIPAddress_Object = MibTableColumn
snmpManagerIPAddress = _SnmpManagerIPAddress_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 2, 1, 2),
    _SnmpManagerIPAddress_Type()
)
snmpManagerIPAddress.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    snmpManagerIPAddress.setStatus("current")


class _SnmpManagerEnable_Type(Integer32):
    """Custom type snmpManagerEnable based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("false", 0),
          ("true", 1))
    )


_SnmpManagerEnable_Type.__name__ = "Integer32"
_SnmpManagerEnable_Object = MibTableColumn
snmpManagerEnable = _SnmpManagerEnable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 2, 1, 3),
    _SnmpManagerEnable_Type()
)
snmpManagerEnable.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    snmpManagerEnable.setStatus("current")
_OutletTable_Object = MibTable
outletTable = _OutletTable_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3)
)
if mibBuilder.loadTexts:
    outletTable.setStatus("current")
_OutletEntry_Object = MibTableRow
outletEntry = _OutletEntry_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3, 1)
)
outletEntry.setIndexNames(
    (0, "IBOOTBAR-MIB", "outletIndex"),
)
if mibBuilder.loadTexts:
    outletEntry.setStatus("current")


class _OutletIndex_Type(Integer32):
    """Custom type outletIndex based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 7),
    )


_OutletIndex_Type.__name__ = "Integer32"
_OutletIndex_Object = MibTableColumn
outletIndex = _OutletIndex_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3, 1, 1),
    _OutletIndex_Type()
)
outletIndex.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    outletIndex.setStatus("current")


class _OutletName_Type(OctetString):
    """Custom type outletName based on OctetString"""
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(20, 20),
    )
    fixedLength = 20


_OutletName_Type.__name__ = "OctetString"
_OutletName_Object = MibTableColumn
outletName = _OutletName_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3, 1, 2),
    _OutletName_Type()
)
outletName.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    outletName.setStatus("current")


class _OutletStatus_Type(Integer32):
    """Custom type outletStatus based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2,
              3,
              4,
              5)
        )
    )
    namedValues = NamedValues(
        *(("off", 0),
          ("on", 1),
          ("reboot", 2),
          ("cycle", 3),
          ("onPending", 4),
          ("cyclePending", 5))
    )


_OutletStatus_Type.__name__ = "Integer32"
_OutletStatus_Object = MibTableColumn
outletStatus = _OutletStatus_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3, 1, 3),
    _OutletStatus_Type()
)
outletStatus.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    outletStatus.setStatus("current")


class _OutletCommand_Type(Integer32):
    """Custom type outletCommand based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2)
        )
    )
    namedValues = NamedValues(
        *(("off", 0),
          ("on", 1),
          ("cycle", 2))
    )


_OutletCommand_Type.__name__ = "Integer32"
_OutletCommand_Object = MibTableColumn
outletCommand = _OutletCommand_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3, 1, 4),
    _OutletCommand_Type()
)
outletCommand.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    outletCommand.setStatus("current")


class _OutletAPStatus_Type(Integer32):
    """Custom type outletAPStatus based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("ok", 0),
          ("triggered", 1))
    )


_OutletAPStatus_Type.__name__ = "Integer32"
_OutletAPStatus_Object = MibTableColumn
outletAPStatus = _OutletAPStatus_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 3, 1, 5),
    _OutletAPStatus_Type()
)
outletAPStatus.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    outletAPStatus.setStatus("current")
_Info_ObjectIdentity = ObjectIdentity
info = _Info_ObjectIdentity(
    (1, 3, 6, 1, 4, 1, 1418, 4, 4)
)
_CurrentLC1_Type = Integer32
_CurrentLC1_Object = MibScalar
currentLC1 = _CurrentLC1_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 4, 1),
    _CurrentLC1_Type()
)
currentLC1.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    currentLC1.setStatus("current")
_CurrentLC2_Type = Integer32
_CurrentLC2_Object = MibScalar
currentLC2 = _CurrentLC2_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 4, 2),
    _CurrentLC2_Type()
)
currentLC2.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    currentLC2.setStatus("current")


class _NumberOfLineCords_Type(Integer32):
    """Custom type numberOfLineCords based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1)
        )
    )
    namedValues = NamedValues(
        *(("oneLineCord", 0),
          ("twoLineCords", 1))
    )


_NumberOfLineCords_Type.__name__ = "Integer32"
_NumberOfLineCords_Object = MibScalar
numberOfLineCords = _NumberOfLineCords_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 4, 3),
    _NumberOfLineCords_Type()
)
numberOfLineCords.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    numberOfLineCords.setStatus("current")


class _EmailStatus_Type(Integer32):
    """Custom type emailStatus based on Integer32"""
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2,
              3,
              4,
              5,
              6)
        )
    )
    namedValues = NamedValues(
        *(("emailSuccess", 0),
          ("emailTimeout", 1),
          ("emailInvalidResponse", 2),
          ("emailDNSFail", 3),
          ("emailAborted", 4),
          ("emailAuthFailed", 5),
          ("errorNotAvail", 6))
    )


_EmailStatus_Type.__name__ = "Integer32"
_EmailStatus_Object = MibScalar
emailStatus = _EmailStatus_Object(
    (1, 3, 6, 1, 4, 1, 1418, 4, 4, 4),
    _EmailStatus_Type()
)
emailStatus.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    emailStatus.setStatus("current")

# Managed Objects groups


# Notification objects

outletChange = NotificationType(
    (1, 3, 6, 1, 4, 1, 1418, 4, 5)
)
outletChange.setObjects(
      *(("IBOOTBAR-MIB", "outletName"),
        ("IBOOTBAR-MIB", "outletStatus"))
)
if mibBuilder.loadTexts:
    outletChange.setStatus(
        "current"
    )

autoPingFailed = NotificationType(
    (1, 3, 6, 1, 4, 1, 1418, 4, 6)
)
autoPingFailed.setObjects(
    ("IBOOTBAR-MIB", "outletAPStatus")
)
if mibBuilder.loadTexts:
    autoPingFailed.setStatus(
        "current"
    )

currentAlarm = NotificationType(
    (1, 3, 6, 1, 4, 1, 1418, 4, 7)
)
currentAlarm.setObjects(
      *(("IBOOTBAR-MIB", "currentLC1"),
        ("IBOOTBAR-MIB", "currentLC2"))
)
if mibBuilder.loadTexts:
    currentAlarm.setStatus(
        "current"
    )

emailError = NotificationType(
    (1, 3, 6, 1, 4, 1, 1418, 4, 8)
)
emailError.setObjects(
    ("IBOOTBAR-MIB", "emailStatus")
)
if mibBuilder.loadTexts:
    emailError.setStatus(
        "current"
    )


# Notifications groups


# Agent capabilities


# Module compliance


# Export all MIB objects to the MIB builder

mibBuilder.exportSymbols(
    "IBOOTBAR-MIB",
    **{"TC1": TC1,
       "dataprobe": dataprobe,
       "iBootBarAgent": iBootBarAgent,
       "systemSettings": systemSettings,
       "deviceName": deviceName,
       "ipMode": ipMode,
       "ipAddress": ipAddress,
       "subnetMask": subnetMask,
       "gateway": gateway,
       "webEnable": webEnable,
       "webPort": webPort,
       "sslEnable": sslEnable,
       "telnetEnable": telnetEnable,
       "telnetPort": telnetPort,
       "updateEnable": updateEnable,
       "cycleTime": cycleTime,
       "delayTime": delayTime,
       "snmpManagerTable": snmpManagerTable,
       "snmpManagerEntry": snmpManagerEntry,
       "snmpManagerIndex": snmpManagerIndex,
       "snmpManagerIPAddress": snmpManagerIPAddress,
       "snmpManagerEnable": snmpManagerEnable,
       "outletTable": outletTable,
       "outletEntry": outletEntry,
       "outletIndex": outletIndex,
       "outletName": outletName,
       "outletStatus": outletStatus,
       "outletCommand": outletCommand,
       "outletAPStatus": outletAPStatus,
       "info": info,
       "currentLC1": currentLC1,
       "currentLC2": currentLC2,
       "numberOfLineCords": numberOfLineCords,
       "emailStatus": emailStatus,
       "outletChange": outletChange,
       "autoPingFailed": autoPingFailed,
       "currentAlarm": currentAlarm,
       "emailError": emailError}
)



#!/usr/bin/env python3
'''OCS agent for dS378 ethernet relay
'''
import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.devantech_dS378.drivers import DS378

PORT_DEFAULT = 17123

LOCK_RELEASE_SEC = 1.
LOCK_RELEASE_TIMEOUT = 10
ACQ_TIMEOUT = 100


class DS378Agent:
    """OCS agent class for dS378 ethernet relay

    Parameters
    ----------
    ip : string
        IP address
    port : int
        Port number
    """

    def __init__(self, agent, ip, port=PORT_DEFAULT):
        self.active = True
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.take_data = False

        self._dev = DS378(ip=ip, port=port)

        self.initialized = False

        agg_params = {'frame_length': 60}
        self.agent.register_feed('relay',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('sampling_frequency', default=0.5, type=float)
    def acq(self, session, params):
        """acq()

        **Process** - Monitor status of the relay.

        Parameters
        ----------
        sampling_frequency : float, optional
            Sampling frequency in Hz, defaults to 0.5 Hz.

        Notes
        -----
        An example of the session data::

            >>> response.session['data']

            {'V_sppl': 11.8,
             'T_int': 30.8,
             'Relay_1': 0,
             'Relay_2': ...,
             'timestamp': 1736541796.779634
            }
        """
        f_sample = params.get('sampling_frequency', 0.5)
        pace_maker = Pacemaker(f_sample)

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn(
                    f'Could not start acq because {self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            self.take_data = True
            session.data = {}
            last_release = time.time()

            while self.take_data:
                # Release lock
                if time.time() - last_release > LOCK_RELEASE_SEC:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=LOCK_RELEASE_TIMEOUT):
                        print(f'Re-acquire failed: {self.lock.job}')
                        return False, 'Could not re-acquire lock.'

                # Data acquisition
                current_time = time.time()
                data = {'timestamp': current_time, 'block_name': 'relay', 'data': {}}

                try:
                    d_status = self._dev.get_status()
                    relay_list = self._dev.get_relays()
                    if session.degraded:
                        self.log.info('Connection re-established.')
                        session.degraded = False
                except ConnectionError:
                    self.log.error('Failed to get data from relay. Check network connection.')
                    session.degraded = True
                    time.sleep(1)
                    continue

                data['data']['V_sppl'] = d_status['V_sppl']
                data['data']['T_int'] = d_status['T_int']
                for i in range(8):
                    data['data'][f'Relay_{i + 1}'] = relay_list[i]

                field_dict = {'V_sppl': d_status['V_sppl'],
                              'T_int': d_status['T_int']}

                for i in range(8):
                    field_dict[f'Relay_{i + 1}'] = relay_list[i]

                session.data.update(field_dict)

                self.agent.publish_to_feed('relay', data)
                session.data.update({'timestamp': current_time})

                pace_maker.sleep()

            self.agent.feeds['relay'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'

        return False, 'acq is not currently running.'

    @ocs_agent.param('relay_number', type=int, check=lambda x: 1 <= x <= 8)
    @ocs_agent.param('on_off', type=int, choices=[0, 1])
    @ocs_agent.param('pulse_time', default=None, type=int, check=lambda x: 0 <= x <= 2**32 - 1)
    def set_relay(self, session, params=None):
        """set_relay(relay_number, on_off, pulse_time=None)

        **Task** - Turns the relay on/off or pulses it.

        Parameters
        ----------
        relay_number : int
            Relay number to manipulate. Values must be in range [1, 8].
        on_off : int
            1 (0) to turn on (off).
        pulse_time : int, optional
            Pulse time in ms. Values must be in range [0, 4294967295].

        Notes
        -----
        This command pulses relay for a given period when ``pulse_time``
        argument is specified, otherwise just turns a relay on or off.

        """
        with self.lock.acquire_timeout(3, job='set_values') as acquired:
            if not acquired:
                self.log.warn('Could not start set_values because '
                              f'{self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            if params.get('pulse_time') is None:
                params['pulse_time'] = 0

            self._dev.set_relay(relay_number=params['relay_number'],
                                on_off=params['on_off'],
                                pulse_time=params['pulse_time'])

        return True, 'Set values'

    def get_relays(self, session, params=None):
        """get_relays()

        **Task** - Returns current relay status.

        Notes
        -----
        The most recent data collected is stored in session.data in the
        structure::

            >>> response.session['data']
            {'Relay_1': 1,
             'Relay_2': ...,
             'timestamp': 1736541796.779634
            }
        """
        with self.lock.acquire_timeout(3, job='get_relays') as acquired:
            if not acquired:
                self.log.warn('Could not start get_relays because '
                              f'{self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            d_status = self._dev.get_relays()
            session.data = {f'Relay_{i + 1}': d_status[i] for i in range(8)}
            session.data.update({'timestamp': time.time()})

        return True, 'Got relay status'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', default=PORT_DEFAULT, type=int,
                        help='Port number for TCP communication.')
    pgroup.add_argument('--ip_address',
                        help='IP address of the device.')

    return parser


def main(args=None):
    """Boot OCS agent"""
    txaio.start_logging(level=os.environ.get('LOGLEVEL', 'info'))

    parser = make_parser()
    args = site_config.parse_args(agent_class='DS378Agent',
                                  parser=parser,
                                  args=args)

    agent_inst, runner = ocs_agent.init_site_agent(args)
    ds_agent = DS378Agent(agent_inst, ip=args.ip_address, port=args.port)

    agent_inst.register_task(
        'set_relay',
        ds_agent.set_relay
    )

    agent_inst.register_task(
        'get_relays',
        ds_agent.get_relays
    )

    agent_inst.register_process(
        'acq',
        ds_agent.acq,
        ds_agent._stop_acq,
        startup=True
    )

    runner.run(agent_inst, auto_reconnect=True)


if __name__ == '__main__':
    main()
